/*
 * GOATTRACKER v1.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <conio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "bme.h"
#include "gconsole.h"
#include "gsound.h"
#include "gsid.h"

#define VK_PAGEUP 33
#define VK_PAGEDOWN 34

#define FORMAT_SID 0
#define FORMAT_PRG 1
#define FORMAT_BIN 2

#define CMD_ARPEGGIO 0
#define CMD_PORTAMENTO 1
#define CMD_SETCUTOFFADD 2
#define CMD_TONEPORTA 3
#define CMD_VIBRATO 4
#define CMD_SETFILTER 5
#define CMD_SETSUSTAIN 6
#define CMD_SETTEMPO 7

#define EDIT_PATTERN 0
#define EDIT_ORDERLIST 1
#define EDIT_INSTRUMENT 2
#define EDIT_NAMES 3

#define MAX_FILES 4096
#define MAX_FILENAME 40
#define MAX_STR 32
#define MAX_INSTR 32
#define MAX_CHN 3
#define MAX_PATT 254
#define MAX_WAVELEN 64
#define MAX_INSTRNAMELEN 16
#define MAX_PATTROWS 80
#define MAX_SONGLEN 255
#define MAX_SONGS 32

#define ENDSONG 0xfe
#define LOOPSONG 0xff

#define ENDPATT 0xff
#define KEYOFF 0x5e
#define REST 0x5f

#define DEFAULTPLAYBUFFERS 10
#define DEFAULTMIXRATE 44100

#define PAGEUPDNREPEAT 4

#define VISIBLEPATTROWS 23
#define VISIBLEORDERLIST 11
#define VISIBLEWAVEROWS 7
#define VISIBLEFILES 19

typedef struct
{
  unsigned char ad;
  unsigned char sr;
  unsigned char pulse;
  unsigned char pulseadd;
  unsigned char pulselimitlow;
  unsigned char pulselimithigh;
  unsigned char filtertype;
  unsigned char wavetableindex;
  unsigned char name[MAX_INSTRNAMELEN];
} INSTR;

typedef struct
{
  char name[MAX_FILENAME];
  int attribute;
} DIRENTRY;

typedef struct
{
  unsigned char opcode;
  int length;
} INSTRUCTION;

typedef struct
{
  unsigned char instrnum;
  unsigned char note;
  unsigned char newnote;
  unsigned char pattptr;
  unsigned char pattnum;
  unsigned char songptr;
  unsigned char freqlo;
  unsigned char freqhi;
  unsigned char wave;
  unsigned char wavetable;
  unsigned char waveinstr;
  unsigned char pulse;
  unsigned char pulsedir;
  unsigned char vibdir;
  unsigned char arpcount;
  unsigned char command;
  unsigned char cmddata;
  unsigned char tick;
  unsigned char tempo;
  unsigned char mute;
} CHN;

CHN chn[MAX_CHN];
INSTR instr[MAX_INSTR];
unsigned char wavetable[MAX_INSTR][MAX_WAVELEN*2];
unsigned char songorder[MAX_SONGS][MAX_CHN][MAX_SONGLEN+1];
unsigned char pattern[MAX_PATT][MAX_PATTROWS*3+3];
unsigned char patterncopybuffer[MAX_PATTROWS*3+3];
INSTR instrcopybuffer;
unsigned char wavecopybuffer[MAX_WAVELEN*2];
int pattlen[MAX_PATT];
int songlen[MAX_SONGS][MAX_CHN];

unsigned char songfilename[MAX_FILENAME];
unsigned char instrfilename[MAX_FILENAME];
unsigned char songfilter[MAX_FILENAME];
unsigned char instrfilter[MAX_FILENAME];
DIRENTRY direntry[MAX_FILES];

unsigned char songname[MAX_STR];
unsigned char authorname[MAX_STR];
unsigned char copyrightname[MAX_STR];

int espos[MAX_CHN];
int eseditpos;
int esview;
int escolumn;
int eschn;
int esnum;
int epnum[MAX_CHN];
int eppos;
int epview;
int epcolumn;
int epchn;
int epoctave = 2;
int einum;
int eipos;
int eicolumn;
int ewpos;
int ewview;
int enpos;
int editmode = EDIT_PATTERN;
int recordmode = 0;
int followplay = 0;
int hexnybble = -1;
int highestusedpattern;
int highestusedinstr;
int scrrep;

unsigned playeradr = 0x1000;
int fileformat = FORMAT_PRG;

unsigned char freqtbllo[] =
 {0x19,0x29,0x3b,0x4e,0x62,0x77,0x8d,0xa5,0xbe,0xd8,0xf4,0x12,
  0x32,0x53,0x76,0x9c,0xc4,0xee,0x1a,0x4a,0x7c,0xb1,0xe9,0x24,
  0x64,0xa6,0xed,0x38,0x88,0xdc,0x35,0x94,0xf8,0x62,0xd2,0x49,
  0xc8,0x4d,0xdb,0x71,0x10,0xb8,0x6b,0x28,0xf0,0xc4,0xa5,0x93,
  0x90,0x9b,0xb6,0xe2,0x20,0x71,0xd6,0x50,0xe0,0x89,0x4a,0x27,
  0x20,0x36,0x6d,0xc5,0x41,0xe2,0xac,0xa0,0xc1,0x12,0x95,0x4e,
  0x40,0x6d,0xda,0x8a,0x82,0xc5,0x59,0x41,0x83,0x25,0x2b,0x9d,
  0x80,0xda,0xb4,0x15,0x04,0x8b,0xb2,0x83,0x07,0x4a,0x57,0xff,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

unsigned char freqtblhi[] =
 {0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x02,
  0x02,0x02,0x02,0x02,0x02,0x02,0x03,0x03,0x03,0x03,0x03,0x04,
  0x04,0x04,0x04,0x05,0x05,0x05,0x06,0x06,0x06,0x07,0x07,0x08,
  0x08,0x09,0x09,0x0a,0x0b,0x0b,0x0c,0x0d,0x0d,0x0e,0x0f,0x10,
  0x11,0x12,0x13,0x14,0x16,0x17,0x18,0x1a,0x1b,0x1d,0x1f,0x21,
  0x23,0x25,0x27,0x29,0x2c,0x2e,0x31,0x34,0x37,0x3b,0x3e,0x42,
  0x46,0x4a,0x4e,0x53,0x58,0x5d,0x63,0x69,0x6f,0x76,0x7d,0x84,
  0x8c,0x94,0x9d,0xa7,0xb1,0xbb,0xc6,0xd2,0xdf,0xec,0xfa,0xff,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

INSTRUCTION asmtable[] = {{0x69,2}, //ADC immediate
                {0x65,2}, //ADC zeropage
                {0x75,2}, //ADC zeropage,X
                {0x6d,3}, //ADC absolute
                {0x7d,3}, //ADC absolute,X
                {0x79,3}, //ADC absolute,Y
                {0x61,2}, //ADC indirect,X
                {0x71,2}, //ADC indirect,Y

                {0x29,2}, //AND immediate
                {0x25,2}, //AND zeropage
                {0x35,2}, //AND zeropage,X
                {0x2d,3}, //AND absolute
                {0x3d,3}, //AND absolute,X
                {0x39,3}, //AND absolute,Y
                {0x21,2}, //AND indirect,X
                {0x31,2}, //AND indirect,Y

                {0x0a,1}, //ASL accumulator
                {0x06,2}, //ASL zeropage
                {0x16,2}, //ASL zeropage,X
                {0x0e,3}, //ASL absolute
                {0x1e,3}, //ASL absolute,X
                {0x90,2}, //BCC
                {0xb0,2}, //BCS
                {0xf0,2}, //BEQ

                {0x24,2}, //BIT zeropage
                {0x2c,3}, //BIT absolute
                {0x30,2}, //BMI
                {0xd0,2}, //BNE
                {0x10,2}, //BPL
                {0x00,1}, //BRK
                {0x50,2}, //BVC
                {0x70,2}, //BVS

                {0x18,1}, //CLC
                {0xd8,1}, //CLD
                {0x58,1}, //CLI
                {0xb8,1}, //CLV
                {0xc9,2}, //CMP immediate
                {0xc5,2}, //CMP zeropage
                {0xd5,2}, //CMP zeropage,X
                {0xcd,3}, //CMP absolute

                {0xdd,3}, //CMP absolute,X
                {0xd9,3}, //CMP absolute,Y
                {0xc1,2}, //CMP indirect,X
                {0xd1,2}, //CMP indirect,Y
                {0xe0,2}, //CPX immediate
                {0xe4,2}, //CPX zeropage
                {0xec,3}, //CPX absolute
                {0xc0,2}, //CPY immediate

                {0xc4,2}, //CPY zeropage
                {0xcc,3}, //CPY absolute
                {0xc6,2}, //DEC zeropage
                {0xd6,2}, //DEC zeropage,X
                {0xce,3}, //DEC absolute
                {0xde,3}, //DEC absolute,X
                {0xca,1}, //DEX
                {0x88,1}, //DEY

                {0x49,2}, //EOR immediate
                {0x45,2}, //EOR zeropage
                {0x55,2}, //EOR zeropage,X
                {0x4d,3}, //EOR absolute
                {0x5d,3}, //EOR absolute,X
                {0x59,3}, //EOR absolute,Y
                {0x41,2}, //EOR indirect,X
                {0x51,2}, //EOR indirect,Y

                {0xe6,2}, //INC zeropage
                {0xf6,2}, //INC zeropage,X
                {0xee,3}, //INC absolute
                {0xfe,3}, //INC absolute,X
                {0xe8,1}, //INX
                {0xc8,1}, //INY
                {0x4c,3}, //JMP absolute
                {0x6c,3}, //JMP indirect

                {0x20,3}, //JSR absolute
                {0xa9,2}, //LDA immediate
                {0xa5,2}, //LDA zeropage
                {0xb5,2}, //LDA zeropage,X
                {0xad,3}, //LDA absolute
                {0xbd,3}, //LDA absolute,X
                {0xb9,3}, //LDA absolute,Y
                {0xa1,2}, //LDA indirect,X

                {0xb1,2}, //LDA indirect,Y
                {0xa2,2}, //LDX immediate
                {0xa6,2}, //LDX zeropage
                {0xb6,2}, //LDX zeropage,Y
                {0xae,3}, //LDX absolute
                {0xbe,3}, //LDX absolute,Y
                {0xa0,2}, //LDY immediate
                {0xa4,2}, //LDY zeropage

                {0xb4,2}, //LDY zeropage,Y
                {0xac,3}, //LDY absolute
                {0xbc,3}, //LDY absolute,Y
                {0x4a,1}, //LSR accumulator
                {0x46,2}, //LSR zeropage
                {0x56,2}, //LSR zeropage,X
                {0x4e,3}, //LSR absolute
                {0x5e,3}, //LSR absolute,X

                {0xea,1}, //NOP
                {0x09,2}, //ORA immediate
                {0x05,2}, //ORA zeropage
                {0x15,2}, //ORA zeropage,X
                {0x0d,3}, //ORA absolute
                {0x1d,3}, //ORA absolute,X
                {0x19,3}, //ORA absolute,Y

                {0x11,2}, //ORA indirect,Y
                {0x48,1}, //PHA
                {0x08,1}, //PHP
                {0x68,1}, //PLA
                {0x28,1}, //PLP
                {0x2a,1}, //ROL accumulator
                {0x26,2}, //ROL zeropage

                {0x36,2}, //ROL zeropage,X
                {0x2e,3}, //ROL absolute
                {0x3e,3}, //ROL absolute,X
                {0x6a,1}, //ROR accumulator
                {0x66,2}, //ROR zeropage
                {0x76,2}, //ROR zeropage,X
                {0x6e,3}, //ROR absolute
                {0x7e,3}, //ROR absolute,X

                {0x40,1}, //RTI
                {0x60,1}, //RTS
                {0xe9,2}, //SBC immediate
                {0xe5,2}, //SBC zeropage
                {0xf5,2}, //SBC zeropage,X
                {0xed,3}, //SBC absolute
                {0xfd,3}, //SBC absolute,X
                {0xf9,3}, //SBC absolute,Y

                {0xe1,2}, //SBC indirect,X
                {0xf1,2}, //SBC indirect,Y
                {0x38,1}, //SEC
                {0xf8,1}, //SED
                {0x78,1}, //SEI
                {0x85,2}, //STA zeropage
                {0x95,2}, //STA zeropage,X
                {0x8d,3}, //STA absolute

                {0x9d,3}, //STA absolute,X
                {0x99,3}, //STA absolute,Y
                {0x81,2}, //STA indirect,X
                {0x91,2}, //STA indirect,Y
                {0x86,2}, //STX zeropage
                {0x96,2}, //STX zeropage,Y
                {0x8e,3}, //STX absolute
                {0x84,2}, //STY zeropage

                {0x94,2}, //STY zeropage,X
                {0x8c,3}, //STY absolute
                {0xaa,1}, //TAX
                {0xa8,1}, //TAY
                {0xba,1}, //TSX
                {0x8a,1}, //TXA
                {0x9a,1}, //TXS
                {0x98,1}, //TYA
                {0xff,0}};


unsigned char volume = 15;
unsigned char filterctrl = 0;
unsigned char filtertype = 0;
unsigned char filtercutoff = 0;
unsigned char filtercutoffadd = 0;
unsigned char songinit;
int playsongnum;

int timemin = 0;
int timesec = 0;
int timeframe = 0;
int cursorflashdelay = 0;
int cursorflash = 0;
int cursorcolortable[] = {1,2,7,2};

unsigned timecounter = 0;
unsigned updates = 0;
unsigned currenttime = 0;
unsigned prevtime = 0;

int key;
int virtualkey;

unsigned char notekeytbl1[] = {'z', 's', 'x', 'd', 'c', 'v',
  'g', 'b', 'h', 'n', 'j', 'm', ',', 'l', '.'};

unsigned char notekeytbl2[] = {'q', '2', 'w', '3', 'e', 'r',
  '5', 't', '6', 'y', '7', 'u', 'i', '9', 'o', '0', 'p'};

unsigned char hexkeytbl[] = {'0', '1', '2', '3', '4', '5', '6', '7',
  '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};

char timechar[] = {':', ' '};

char *notename[] =
 {"C-0", "C#0", "D-0", "D#0", "E-0", "F-0", "F#0", "G-0", "G#0", "A-0", "A#0", "B-0",
  "C-1", "C#1", "D-1", "D#1", "E-1", "F-1", "F#1", "G-1", "G#1", "A-1", "A#1", "B-1",
  "C-2", "C#2", "D-2", "D#2", "E-2", "F-2", "F#2", "G-2", "G#2", "A-2", "A#2", "B-2",
  "C-3", "C#3", "D-3", "D#3", "E-3", "F-3", "F#3", "G-3", "G#3", "A-3", "A#3", "B-3",
  "C-4", "C#4", "D-4", "D#4", "E-4", "F-4", "F#4", "G-4", "G#4", "A-4", "A#4", "B-4",
  "C-5", "C#5", "D-5", "D#5", "E-5", "F-5", "F#5", "G-5", "G#5", "A-5", "A#5", "B-5",
  "C-6", "C#6", "D-6", "D#6", "E-6", "F-6", "F#6", "G-6", "G#6", "A-6", "A#6", "B-6",
  "C-7", "C#7", "D-7", "D#7", "E-7", "F-7", "F#7", "G-7", "G#7", "A-7", "===", "---"};

char textbuffer[80];
char exitprogram = 0;
char *programname = "GOATTRACKER v1.0 by Lasse rni (reSID by Dag Lem)";

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow);
void getkey(void);
void waitkey(void);
void waitkey2(void);
void relocator(void);
int testoverlap(int area1start, int area1size, int area2start, int area2size);
int packpattern(unsigned char *dest, unsigned char *src, int rows);
void handle_int(void);
void printmainscreen(void);
void docommand(void);
void orderlistcommands(void);
void patterncommands(void);
void instrumentcommands(void);
void namecommands(void);
void display_update(void);
void editstring(char *buffer, int maxlength);
int fileselector(char *name, char *filter, char *title, int initialmode);
void clearsong(int cs, int cp, int ci, int cn);
void countpatternlengths(void);
void countthispattern(void);
void printpatterns(void);
void playroutine(void);
void freqadd(int c, CHN *cptr, unsigned char speed);
void freqsub(int c, CHN *cptr, unsigned char speed);
unsigned char swapnybbles(unsigned char n);

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
  unsigned b = DEFAULTPLAYBUFFERS;
  unsigned mr = DEFAULTMIXRATE;
  unsigned sm = GFX_SCANLINES;
  unsigned writer = 0;
  unsigned hardsid = 0;
  char filename[MAX_PATH];
  FILE *configfile;
  unsigned c;
  int info = 0;

  win_fullscreen = 1;

  GetModuleFileName(NULL, (TCHAR *)filename, MAX_PATH);
  io_opendatafile(filename);

  /* Set signal handler */
  signal(SIGINT, handle_int);

  /* Get configuration info */
  filename[strlen(filename)-3] = 'c';
  filename[strlen(filename)-2] = 'f';
  filename[strlen(filename)-1] = 'g';
  configfile = fopen(filename, "rt");
  if (configfile)
  {
        fscanf(configfile, "%d", &b);
        fscanf(configfile, "%d", &mr);
        fscanf(configfile, "%d", &sm);
        fscanf(configfile, "%d", &win_fullscreen);
        fscanf(configfile, "%d", &hardsid);
        fclose(configfile);
  }

  /* Open application window */
  win_openwindow("GoatTracker");

  /* Scan command line */
  strupr(szCmdLine);
  for (c = 0; c < strlen(szCmdLine); c++)
  {
    if ((szCmdLine[c] == '-') || (szCmdLine[c] == '/'))
    {
      switch(szCmdLine[c+1])
      {
        case '?':
        win_messagebox("Usage: GOATTRK [options]\n\n"
               "Options:\n"
               "/Bxx Set number of playback buffers (20ms each) DEFAULT=10\n"
               "/Mxx Set mixing rate DEFAULT=44100\n"
               "/Sxx Set screenmode (0 = singlesize 1 = scanlines 2 = doublesize)\n"
               "/Fxx Set fullscreenmode (0 = off 1 = on)\n"
               "/Hxx Use HardSID (0 = off, 1 = HardSID ID0 2 = HardSID ID1 etc.)\n"
               "/W   Write sound output to a file SIDAUDIO.RAW\n"
               "/?   Show this info again\n\n"
               "Settings will be stored into a file called GOATTRK.CFG in the same\n"
               "directory as the executable file.\n");
        info = 1;
        break;

        case 'B':
        sscanf(&szCmdLine[c+2], "%u", &b);
        break;

        case 'M':
        sscanf(&szCmdLine[c+2], "%u", &mr);
        break;

        case 'H':
        sscanf(&szCmdLine[c+2], "%u", &hardsid);
        break;

        case 'W':
        writer = 1;
        break;

        case 'S':
        sscanf(&szCmdLine[c+2], "%u", &sm);
        if (sm > 2) sm = 2;
        break;

        case 'F':
        sscanf(&szCmdLine[c+2], "%u", &win_fullscreen);
        if (win_fullscreen < 0) win_fullscreen = 0;
        if (win_fullscreen > 1) win_fullscreen = 1;
        break;
      }
    }
    if (info) return 0;
  }

  if (!sound_init(b, mr, writer, hardsid))
  {
    win_messagebox("Sound init failed!");
    return 1;
  }
  atexit(sound_uninit);

  if (!initscreen(sm))
  {
    win_messagebox("Graphics init failed!");
    return 1;
  }
  atexit(closescreen);

  memset(songfilename, 0, sizeof songfilename);
  memset(instrfilename, 0, sizeof instrfilename);
  strcpy(songfilter, "*.sng");
  strcpy(instrfilter, "*.ins");
  memset(chn, 0, sizeof chn);
  clearsong(1,1,1,1);
  printmainscreen();

  while (!exitprogram)
  {
    waitkey();
    docommand();
  }

  /* Store configuration to config file */
  configfile = fopen(filename, "wt");
  if (configfile)
  {
        fprintf(configfile, "%d %d %d %d %d", b, mr, sm, win_fullscreen, hardsid);
        fclose(configfile);
  }

  END:
  return 0;
}

void display_update(void)
{
  if (cursorflashdelay >= 6)
  {
    cursorflashdelay %= 6;
    cursorflash++;
    cursorflash &= 3;
    printpatterns();
    fliptoscreen();
  }
  else
  {
    if (virtualkey)
    {
      printpatterns();
      fliptoscreen();
    }
  }
}

void docommand(void)
{
  int c;

  hexnybble = -1;
  for (c = 0; c < 16; c++)
  {
    if (tolower(key) == hexkeytbl[c]) hexnybble = c;
  }

  /* Mode-specific commands */
  switch(editmode)
  {
    case EDIT_ORDERLIST:
    orderlistcommands();
    break;

    case EDIT_INSTRUMENT:
    instrumentcommands();
    break;

    case EDIT_PATTERN:
    patterncommands();
    break;

    case EDIT_NAMES:
    namecommands();
    break;
  }

  /* General commands */
  switch(key)
  {
    case '*':
    if ((eipos != 7) || (editmode != EDIT_INSTRUMENT))
    {
      if (editmode != EDIT_NAMES)
      {
        if (epoctave < 7) epoctave++;
      }
    }
    break;

    case '/':
    if ((eipos != 7) || (editmode != EDIT_INSTRUMENT))
    {
      if (editmode != EDIT_NAMES)
      {
        if (epoctave > 0) epoctave--;
      }
    }
    break;

    case '-':
    if ((eipos != 7) || (editmode != EDIT_INSTRUMENT))
    {
      if (editmode != EDIT_NAMES)
      {
        einum--;
        if (einum < 0) einum = 0;
        ewpos = 0;
        ewview = 0;
      }
    }
    break;

    case '+':
    if ((eipos != 7) || (editmode != EDIT_INSTRUMENT))
    {
      if (editmode != EDIT_NAMES)
      {
        einum++;
        if (einum >= MAX_INSTR) einum = MAX_INSTR - 1;
        ewpos = 0;
        ewview = 0;
      }
    }
    break;
  }
  if (win_quitted) exitprogram = 1;
  switch(virtualkey)
  {
    case VK_ESCAPE:
    if ((!win_keytable[KEY_LEFTSHIFT]) && (!win_keytable[KEY_RIGHTSHIFT]))
    {
      printtextc(29, 15, "Really Quit (y/n)?");
      waitkey();
      printblank(20, 29, 40);
      if ((key == 'y') || (key == 'Y')) exitprogram = 1;
      break;
    }
    else
    {
      int cs = 0;
      int cp = 0;
      int ci = 0;
      int cn = 0;

      printtextc(29, 15, "Clear orderlists (y/n)?");
      waitkey();
      printblank(20, 29, 40);
      if ((key == 'y') || (key == 'Y')) cs = 1;

      printtextc(29, 15, "Clear patterns (y/n)?");
      waitkey();
      printblank(20, 29, 40);
      if ((key == 'y') || (key == 'Y')) cp = 1;

      printtextc(29, 15, "Clear instruments (y/n)?");
      waitkey();
      printblank(20, 29, 40);
      if ((key == 'y') || (key == 'Y')) ci = 1;

      printtextc(29, 15, "Clear songname (y/n)?");
      waitkey();
      printblank(20, 29, 40);
      if ((key == 'y') || (key == 'Y')) cn = 1;
      clearsong(cs, cp, ci, cn);
      break;
    }


    case VK_F12:
    clearscreen();
    printtextc(0, 15,   "GOATTRACKER online help 1/2: Keyboard (press any key for next page)");
    printtext(0, 2, 15, "GENERAL KEYS");
    printtext(0, 3, 7,  "TAB Switch between editing modes");
    printtext(0, 4, 7,  "F1  Play from beginning    (With SHIFT=");
    printtext(0, 5, 7,  "F2  Play from current pos. follow play)");
    printtext(0, 6, 7,  "F3  Stop playing/silence all sounds");
    printtext(0, 7, 7,  "F5  Copy pattern/instrument");
    printtext(0, 8, 7,  "F6  Paste pattern/instrument");
    printtext(0, 9, 7, "F9  Pack, relocate & save PRG,SID etc.");
    printtext(0, 10, 7, "F10 Load song/instrument");
    printtext(0, 11, 7, "F11 Save song/instrument");
    printtext(0, 12, 7, "F12 This screen");
    printtext(0, 13, 7, "SHIFT+ESC Clear musicdata");
    printtext(0, 14, 7, "ESC Exit program");
    printtext(40, 2, 15, "PATTERN EDIT MODE");
    printtext(40, 3,  7, "Enter notes like on piano keyboard");
    printtext(40, 4,  7, "0-9 & A-F to enter commands");
    printtext(40, 5,  7, "SPC Switch between jam/editmode");
    printtext(40, 6,  7, "< (or CAPSLOCK) Insert keyoff");
    printtext(40, 7,  7, "< Insert rest");
    printtext(40, 8,  7, "INS Insert row (Press on the endmark");
    printtext(40, 9,  7, "DEL Delete row to change pat.length)");
    printtext(40, 10,  7, "- + Select instrument");
    printtext(40, 11,  7, "/ * Select octave");
    printtext(40, 12,  7, "< > Select pattern");
    printtext(40, 13,  7, "F4  Mute current channel");
    printtext(40, 14,  7, "F7  Transpose pattern one halfstep down");
    printtext(40, 15,  7, "F8  Transpose pattern one halfstep up");
    printtext(40, 17, 15, "SONG EDIT MODE");
    printtext(40, 18,  7, "0-9 & A-F to enter pattern numbers");
    printtext(40, 19,  7, "SPC Set start position for F2 key");
    printtext(40, 20,  7, "< Go to pattern under cursor");
    printtext(40, 21,  7, "< > Select subtune");
    printtext(40, 22,  7, "R S Change endmark to RESTART or STOP");
    printtext(40, 23,  7, "INS Insert column (Press on the endmark");
    printtext(40, 24,  7, "DEL Delete column to change songlength)");
    printtext(0, 16, 15, "INSTRUMENT EDIT MODE");
    printtext(0, 17, 7, "0-9 & A-F to enter parameters");
    printtext(0, 18, 7, "SPC Play test note");
    printtext(0, 19, 7, "< Silence test note");
    printtext(0, 20, 7, "- + Select instrument");
    printtext(0, 21, 7, "/ * Select octave");
    printtext(0, 23, 15,"SONGNAME EDIT MODE");
    printtext(0, 24, 7,"Use cursor UP/DOWN to move between");
    printtext(0, 25, 7,"song, author & copyright strings,");
    printtext(0, 26, 7,"other keys to edit them.");
    fliptoscreen();
    waitkey2();
    if (virtualkey == VK_ESCAPE) goto EXITHELP;

    clearscreen();
    printtextc(0, 15,   "GOATTRACKER online help page 2/2: Commands (press any key to return to editor)");
    printtext(0, 2, 15, "COMMAND 0XY: ");
    printtext(13, 2, 7, "Arpeggio. Arpeggiates with the root note, then root note + X      ");
    printtext(13,3, 7, "halftones and then root note + Y halftones. If the X value        ");
    printtext(13,4, 7, "is 8 or higher, the arpeggio will be at half speed (8 will        ");
    printtext(13,5, 7, "be subtracted from the value of X to get the amount of halftones) ");
    printtext(0, 6, 15, "COMMAND 1XY: ");
    printtext(13, 6, 7, "Portamento. Raises the pitch with (XY and 7F) * 2 each tick.      ");
    printtext(13,7, 7, "Highest bit of XY decides direction: 0 is up and 1 is down.       ");
    printtext(0, 8,15, "COMMAND 2XY: ");
    printtext(13,8, 7, "Set filter cutoff speed. The cutoff speed will be added to the    ");
    printtext(13,9, 7, "cutoff frequency each tick, until stopped with a 200 command.     ");
    printtext(0,10,15, "COMMAND 3XY: ");
    printtext(13,10, 7, "Toneportamento. Raises or lowers pitch until target note has been ");
    printtext(13,11, 7, "reached. Highest bit of XY decides direction of toneportamento    ");
    printtext(13,12, 7, "(this is an optimization for the C64 playroutine, to keep it fast:");
    printtext(13,13, 7, "you must know the direction yourself) and the actual speed used   ");
    printtext(13,14, 7, "is (XY and 7F) * 2. By specifying the 'wrong' direction you can   ");
    printtext(13,15, 7, "get infinitely fast toneportamentos for 'tie note' effect.        ");
    printtext(0,16,15, "COMMAND 4XY: ");
    printtext(13,16,7,  "Vibrato. X determines how fast the direction will change (E is    ");
    printtext(13,17, 7, "slowest, lowest bit doesn't matter) and Y*16+X is the amount of   ");
    printtext(13,18, 7, "pitch change on each tick.                                        ");
    printtext(0,19,15, "COMMAND 5XY: ");
    printtext(13,19,7, "Set filter parameters. X is the filter resonance and Y is the     ");
    printtext(13,20, 7, "bitmask of channels to be filtered.                               ");
    printtext(13,21, 7, "  Bitvalue 01 = Channel 1                                         ");
    printtext(13,22, 7, "  Bitvalue 02 = Channel 2                                         ");
    printtext(13,23, 7, "  Bitvalue 04 = Channel 3                                         ");
    printtext(0,24,15, "COMMAND 6XY: ");
    printtext(13,24,7, "Set sustain/release. Sets the channel's sustain/release register  ");
    printtext(13,25, 7, "to XY.                                                            ");
    printtext(0,26,15, "COMMAND 7XY: ");
    printtext(13,26,7, "Set tempo. If highest bit is 1 (values 80-FF), tempo (XY and 7F)  ");
    printtext(13, 27,7, "will be set on current channel only, otherwise on all channels.   ");
    fliptoscreen();
    waitkey2();
    EXITHELP:
    printmainscreen();
    break;

    case VK_F1:
    songinit = 0x01;
    if ((win_keytable[KEY_LEFTSHIFT]) || (win_keytable[KEY_RIGHTSHIFT]))
      followplay = 1;
    else
      followplay = 0;
    break;

    case VK_F2:
    songinit = 0x02;
    if ((win_keytable[KEY_LEFTSHIFT]) || (win_keytable[KEY_RIGHTSHIFT]))
      followplay = 1;
    else
      followplay = 0;
    break;

    case VK_F3:
    songinit = 0x03;
    break;

    case VK_F9:
    relocator();
    break;

    case VK_F10:
    if (editmode != EDIT_INSTRUMENT)
    {
      if (fileselector(songfilename, songfilter, "LOAD SONG", 0))
      {
        char ident[4];
        int handle;

        handle = open(songfilename, O_RDONLY | O_BINARY);
        if (handle != -1)
        {
          read(handle, ident, 4);
          if (!memcmp(ident, "GTS!", 4))
          {
            int d;
            unsigned char length;
            unsigned char amount;
            int loadbytes;
            clearsong(1,1,1,1);

            /* Read infotexts */
            read(handle, songname, sizeof songname);
            read(handle, authorname, sizeof authorname);
            read(handle, copyrightname, sizeof copyrightname);

            /* Read songorderlists */
            read(handle, &amount, sizeof amount);
            for (d = 0; d < amount; d++)
            {
              for (c = 0; c < MAX_CHN; c++)
              {
                read(handle, &length, sizeof length);
                loadbytes = length;
                loadbytes++;
                read(handle, songorder[d][c], loadbytes);
              }
            }
            /* Read instruments */
            for (c = 1; c < MAX_INSTR; c++)
            {
              read(handle, &instr[c], sizeof(INSTR));
              loadbytes = instr[c].wavetableindex;
              instr[c].wavetableindex = 1;
              read(handle, wavetable[c], loadbytes);
            }
            /* Read patterns */
            read(handle, &amount, sizeof amount);
            for (c = 0; c < amount; c++)
            {
              read(handle, &length, sizeof length);
              read(handle, pattern[c], length);
            }
            countpatternlengths();
          }
          close(handle);
        }
      }
    }
    else
    {
      if (einum)
      {
        if (fileselector(instrfilename, instrfilter, "LOAD INSTRUMENT", 0))
        {
          char ident[4];
          int handle;

          handle = open(instrfilename, O_RDONLY | O_BINARY);
          if (handle != -1)
          {
            read(handle, ident, 4);
            if (!memcmp(ident, "GTI!", 4))
            {
              int d;
              int loadbytes;

              read(handle, &instr[einum], sizeof(INSTR));
              loadbytes = instr[einum].wavetableindex;
              instr[einum].wavetableindex = 1;
              memset(wavetable[einum], 0, MAX_WAVELEN*2);
              read(handle, wavetable[einum], loadbytes);
            }
            close(handle);
          }
        }
      }
    }
    break;

    case VK_F11:
    if (editmode != EDIT_INSTRUMENT)
    {
      if (fileselector(songfilename, songfilter, "SAVE SONG", 2))
      {
        char ident[] = {'G', 'T', 'S', '!'};
        int handle;

        handle = open(songfilename, O_RDWR | O_BINARY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
        if (handle != -1)
        {
          int d;
          unsigned char length;
          unsigned char amount;
          int writebytes;
          write(handle, ident, 4);

          countpatternlengths();

          /* Write infotexts */
          write(handle, songname, sizeof songname);
          write(handle, authorname, sizeof authorname);
          write(handle, copyrightname, sizeof copyrightname);

          /* Determine amount of songs to be saved */
          c = 0;
          for (;;)
          {
            if (c == MAX_SONGS) break;
            if ((!songlen[c][0])||
               (!songlen[c][1])||
               (!songlen[c][2])) break;
            c++;
          }
          amount = c;

          write(handle, &amount, sizeof amount);
          /* Write songorderlists */
          for (d = 0; d < amount; d++)
          {
            for (c = 0; c < MAX_CHN; c++)
            {
              length = songlen[d][c];
              write(handle, &length, sizeof length);
              writebytes = length;
              writebytes++;
              write(handle, songorder[d][c], writebytes);
            }
          }
          /* Write instruments */
          for (c = 1; c < MAX_INSTR; c++)
          {
            instr[c].wavetableindex = 2;
            for (d = 0; d < MAX_WAVELEN*2; d+= 2)
            {
              if (wavetable[c][d] == 0xff) break;
              instr[c].wavetableindex += 2;
            }
            write(handle, &instr[c], sizeof(INSTR));
            write(handle, wavetable[c], instr[c].wavetableindex);
            instr[c].wavetableindex = 1;
          }
          /* Write patterns */
          amount = highestusedpattern + 1;
          write(handle, &amount, sizeof amount);
          for (c = 0; c < amount; c++)
          {
            length = pattlen[c]*3+3;
            write(handle, &length, sizeof length);
            write(handle, pattern[c], length);
          }
          close(handle);
        }
      }
    }
    else
    {
      if (einum)
      {
        if (fileselector(instrfilename, instrfilter, "SAVE INSTRUMENT", 2))
        {
          char ident[] = {'G', 'T', 'I', '!'};
          int handle;

          handle = open(instrfilename, O_RDWR | O_BINARY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
          if (handle != -1)
          {
            int d;
            int writebytes;
            write(handle, ident, 4);

            /* Write instrument */
            instr[einum].wavetableindex = 2;
            for (d = 0; d < MAX_WAVELEN*2; d+= 2)
            {
              if (wavetable[einum][d] == 0xff) break;
              instr[einum].wavetableindex += 2;
            }
            write(handle, &instr[einum], sizeof(INSTR));
            write(handle, wavetable[einum], instr[einum].wavetableindex);
            instr[einum].wavetableindex = 1;
            close(handle);
          }
        }
      }
    }
    break;
  }
}

void orderlistcommands(void)
{
  int c;

  if (hexnybble >= 0)
  {
    if (eseditpos < songlen[esnum][eschn])
    {
      switch(escolumn)
      {
        case 0:
        songorder[esnum][eschn][eseditpos] &= 0x0f;
        songorder[esnum][eschn][eseditpos] |= hexnybble << 4;
        if (songorder[esnum][eschn][eseditpos] >= MAX_PATT)
          songorder[esnum][eschn][eseditpos] = MAX_PATT - 1;
        break;

        case 1:
        songorder[esnum][eschn][eseditpos] &= 0xf0;
        songorder[esnum][eschn][eseditpos] |= hexnybble;
        if (songorder[esnum][eschn][eseditpos] >= MAX_PATT)
          songorder[esnum][eschn][eseditpos] = MAX_PATT - 1;
        break;
      }
      escolumn++;
      if (eseditpos >= songlen[esnum][eschn]) escolumn = 0;
      if (escolumn > 1)
      {
        escolumn = 0;
        if (eseditpos < songlen[esnum][eschn])
        {
          eseditpos++;
        }
      }
    }
  }

  switch(key)
  {
    case '>':
    case ')':
    case ']':
    esnum++;
    if (esnum >= MAX_SONGS) esnum = MAX_SONGS - 1;
    for (c = 0; c < MAX_CHN; c++)
    {
      espos[c] = 0;
      if (songlen[esnum][c])
      {
        eppos = 0;
        epview = - VISIBLEPATTROWS / 2;
        epnum[c] = songorder[esnum][c][0];
      }
    }
    eseditpos = 0;
    esview = 0;
    break;

    case '<':
    case '(':
    case '[':
    esnum--;
    if (esnum < 0) esnum = 0;
    for (c = 0; c < MAX_CHN; c++)
    {
      espos[c] = 0;
      if (songlen[esnum][c])
      {
        eppos = 0;
        epview = - VISIBLEPATTROWS / 2;
        epnum[c] = songorder[esnum][c][0];
      }
    }
    eseditpos = 0;
    esview = 0;
    break;
  }
  switch(virtualkey)
  {
    case VK_TAB:
    editmode++;
    break;

    case VK_SPACE:
    if (eseditpos < songlen[esnum][eschn])
    {
      espos[eschn] = eseditpos;
    }
    break;

    case VK_RETURN:
    if (eseditpos < songlen[esnum][eschn])
    {
      if ((!win_keytable[KEY_LEFTSHIFT]) && (!win_keytable[KEY_RIGHTSHIFT]))
      {
        epnum[eschn] = songorder[esnum][eschn][eseditpos];
      }
      else
      {
        if (eseditpos < songlen[esnum][0]) epnum[0] = songorder[esnum][0][eseditpos];
        if (eseditpos < songlen[esnum][1]) epnum[1] = songorder[esnum][1][eseditpos];
        if (eseditpos < songlen[esnum][2]) epnum[2] = songorder[esnum][2][eseditpos];
      }
    }
    epchn = eschn;
    epcolumn = 0;
    eppos = 0;
    epview = - VISIBLEPATTROWS / 2;
    editmode = EDIT_PATTERN;
    break;

    case VK_DELETE:
    if (!songinit) songinit = 0x03;
    if ((songlen[esnum][eschn] - eseditpos)-1 >= 0)
    {
      memmove(&songorder[esnum][eschn][eseditpos],
        &songorder[esnum][eschn][eseditpos+1],
        (songlen[esnum][eschn] - eseditpos)-1);
      songorder[esnum][eschn][songlen[esnum][eschn]-1] = 0x00;
    }
    else
    {
      if (eseditpos == songlen[esnum][eschn])
      {
        if (songlen[esnum][eschn] > 0)
        {
          songorder[esnum][eschn][songlen[esnum][eschn]-1] =
            songorder[esnum][eschn][songlen[esnum][eschn]];
          countthispattern();
          eseditpos = songlen[esnum][eschn];
        }
      }
    }
    break;

    case VK_INSERT:
    if (!songinit) songinit = 0x03;
    if ((songlen[esnum][eschn] - eseditpos)-1 >= 0)
    {
      memmove(&songorder[esnum][eschn][eseditpos+1],
        &songorder[esnum][eschn][eseditpos],
        (songlen[esnum][eschn] - eseditpos)-1);
      songorder[esnum][eschn][eseditpos] = 0x00;
    }
    else
    {
      if (eseditpos == songlen[esnum][eschn])
      {
        if (songlen[esnum][eschn] < MAX_SONGLEN)
        {
          songorder[esnum][eschn][eseditpos+1] =
            songorder[esnum][eschn][eseditpos];
          songorder[esnum][eschn][eseditpos] = 0x00;
          countthispattern();
          eseditpos = songlen[esnum][eschn];
        }
      }
    }
    break;
    case 'R':
    if (eseditpos == songlen[esnum][eschn])
    {
      if (!songinit) songinit = 0x03;
      songorder[esnum][eschn][eseditpos] = LOOPSONG;
    }
    break;

    case 'S':
    if (eseditpos == songlen[esnum][eschn])
    {
      if (!songinit) songinit = 0x03;
      songorder[esnum][eschn][eseditpos] = ENDSONG;
    }
    break;

    case VK_RIGHT:
    escolumn++;
    if (eseditpos >= songlen[esnum][eschn]) escolumn = 0;
    if (escolumn > 1)
    {
      escolumn = 0;
      if (eseditpos < songlen[esnum][eschn])
      {
        eseditpos++;
      }
    }
    break;

    case VK_PAGEDOWN:
    for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
    {
      escolumn++;
      if (eseditpos >= songlen[esnum][eschn]) escolumn = 0;
      if (escolumn > 1)
      {
        escolumn = 0;
        if (eseditpos < songlen[esnum][eschn])
        {
          eseditpos++;
        }
      }
    }
    break;

    case VK_HOME:
    eseditpos = 0;
    escolumn = 0;
    break;

    case VK_END:
    eseditpos = songlen[esnum][eschn];
    escolumn = 0;
    break;

    case VK_PAGEUP:
    for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
    {
      escolumn--;
      if (escolumn < 0)
      {
        if (eseditpos > 0)
        {
          eseditpos--;
          escolumn = 1;
        }
        else escolumn = 0;
      }
    }
    break;

    case VK_LEFT:
    escolumn--;
    if (escolumn < 0)
    {
      if (eseditpos > 0)
      {
        eseditpos--;
        escolumn = 1;
      }
      else escolumn = 0;
    }
    break;

    case VK_UP:
    eschn--;
    if (eschn < 0) eschn = MAX_CHN - 1;
    if (eseditpos > songlen[esnum][eschn])
    {
      eseditpos = songlen[esnum][eschn];
      escolumn = 0;
    }
    break;

    case VK_DOWN:
    eschn++;
    if (eschn >= MAX_CHN) eschn = 0;
    if (eseditpos > songlen[esnum][eschn])
    {
      eseditpos = songlen[esnum][eschn];
      escolumn = 0;
    }
    break;
  }
  if (eseditpos - esview < 0)
  {
    esview = eseditpos;
  }
  if (eseditpos - esview >= VISIBLEORDERLIST)
  {
    esview = eseditpos - VISIBLEORDERLIST + 1;
  }
}

void instrumentcommands(void)
{
  switch(virtualkey)
  {
    case VK_TAB:
    editmode++;
    break;

    case VK_F5:
    if (einum)
    {
      memcpy(&instrcopybuffer, &instr[einum], sizeof(INSTR));
      memcpy(wavecopybuffer, wavetable[einum], MAX_WAVELEN*2);
    }
    break;

    case VK_F6:
    if (einum)
    {
      memcpy(&instr[einum], &instrcopybuffer, sizeof(INSTR));
      memcpy(wavetable[einum], wavecopybuffer, MAX_WAVELEN*2);
    }
    break;

    case VK_INSERT:
    if ((eicolumn >= 2) && (ewpos < (MAX_WAVELEN-1)))
    {
      memmove(&wavetable[einum][ewpos*2+2],
        &wavetable[einum][ewpos*2],
        (MAX_WAVELEN-1 - ewpos)*2-2);
      wavetable[einum][ewpos*2] = 0x00;
      wavetable[einum][ewpos*2 + 1] = 0x00;
    }
    break;

    case VK_DELETE:
    if ((eicolumn >= 2) && (ewpos < (MAX_WAVELEN-1)))
    {
      memmove(&wavetable[einum][ewpos*2],
        &wavetable[einum][ewpos*2+2],
        (MAX_WAVELEN-1 - ewpos)*2-2);
      wavetable[einum][(MAX_WAVELEN-2)*2] = 0x00;
      wavetable[einum][(MAX_WAVELEN-2)*2 + 1] = 0x00;
    }
    break;

    case VK_RIGHT:
    if (eipos != 7)
    {
      eicolumn++;
      if (eicolumn > 5) eicolumn = 0;
    }
    break;

    case VK_LEFT:
    if (eipos != 7)
    {
      eicolumn--;
      if (eicolumn < 0) eicolumn = 5;
    }
    break;

    case VK_PAGEDOWN:
    for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
    {
      if (eicolumn < 2)
      {
        eipos++;
        if (eipos > 7) eipos = 0;
      }
      else
      {
        if (ewpos < MAX_WAVELEN-1)
        {
          ewpos++;
        }
      }
    }
    break;

    case VK_DOWN:
    if (eicolumn < 2)
    {
      eipos++;
      if (eipos > 7) eipos = 0;
    }
    else
    {
      if (ewpos < MAX_WAVELEN-1)
      {
        ewpos++;
      }
    }
    break;

    case VK_PAGEUP:
    for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
    {
      if (eicolumn < 2)
      {
        eipos--;
        if (eipos < 0) eipos = 7;
      }
      else
      {
        if (ewpos > 0)
        {
          ewpos--;
        }
      }
    }
    break;

    case VK_UP:
    if (eicolumn < 2)
    {
      eipos--;
      if (eipos < 0) eipos = 7;
    }
    else
    {
      if (ewpos > 0)
      {
        ewpos--;
      }
    }
    break;

    case VK_SPACE:
    if ((eipos != 7) && (songinit == 0x80))
    {
      sidreg[0x4 + epchn * 7] = chn[epchn].wave & 0xfe; /* Hard restart */
      sidreg[0x5 + epchn * 7] = 0;
      sidreg[0x6 + epchn * 7] = 0;
      chn[epchn].newnote = 0xfe; /* Trigger note delayed by two updates */
      chn[epchn].instrnum = einum;
      chn[epchn].note = epoctave * 12;
    }
    break;

    case VK_RETURN:
    if ((eipos != 7) && (songinit == 0x80))
    {
      chn[epchn].wave &= 0xfe;
      sidreg[0x4 + epchn * 7] = chn[epchn].wave; /* Gate off */
    }
    break;
  }
  if ((eipos == 7) && (einum)) editstring(instr[einum].name, MAX_INSTRNAMELEN);
  if ((hexnybble >= 0) && (eipos != 7) && (einum))
  {
    unsigned char *ptr = &instr[einum].ad;
    ptr += eipos;

    switch(eicolumn)
    {
      case 0:
      if ((eipos < 2) || (eipos > 5))
      {
        *ptr &= 0x0f;
        *ptr |= hexnybble << 4;
      }
      else
      {
        *ptr &= 0xf0;
        *ptr |= hexnybble;
      }
      eicolumn++;
      if (eicolumn > 1)
      {
        eicolumn = 0;
        eipos++;
      }
      break;

      case 1:
      if ((eipos < 2) || (eipos > 5))
      {
        *ptr &= 0xf0;
        *ptr |= hexnybble;
      }
      else
      {
        *ptr &= 0x0f;
        *ptr |= hexnybble << 4;
      }
      eicolumn++;
      if (eicolumn > 1)
      {
        eicolumn = 0;
        eipos++;
      }
      break;

      case 2:
      if (ewpos < MAX_WAVELEN-1)
      {
        wavetable[einum][ewpos*2] &= 0x0f;
        wavetable[einum][ewpos*2] |= hexnybble << 4;
        eicolumn++;
        if (eicolumn > 5)
        {
          eicolumn = 2;
          ewpos++;
        }
      }
      break;

      case 3:
      if (ewpos < MAX_WAVELEN-1)
      {
        wavetable[einum][ewpos*2] &= 0xf0;
        wavetable[einum][ewpos*2] |= hexnybble;
        eicolumn++;
        if (eicolumn > 5)
        {
          eicolumn = 2;
          ewpos++;
        }
      }
      break;

      case 4:
      if (ewpos < MAX_WAVELEN-1)
      {
        wavetable[einum][ewpos*2+1] &= 0x0f;
        wavetable[einum][ewpos*2+1] |= hexnybble << 4;
        eicolumn++;
        if (eicolumn > 5)
        {
          eicolumn = 2;
          ewpos++;
        }
      }
      break;

      case 5:
      if (ewpos < MAX_WAVELEN-1)
      {
        wavetable[einum][ewpos*2+1] &= 0xf0;
        wavetable[einum][ewpos*2+1] |= hexnybble;
        eicolumn++;
        if (eicolumn > 5)
        {
          eicolumn = 2;
          ewpos++;
        }
      }
      break;
    }
  }

  if (ewpos - ewview < 0)
  {
    ewview = ewpos;
  }
  if (ewpos - ewview >= VISIBLEWAVEROWS)
  {
    ewview = ewpos - VISIBLEWAVEROWS + 1;
  }

  /* Make sure there aren't illegal parameters */
  instr[einum].pulseadd &= 0xf0;
  instr[einum].pulselimitlow &= 0x0f;
  instr[einum].pulselimithigh &= 0x0f;
}

void namecommands(void)
{
  switch(virtualkey)
  {
    case VK_TAB:
    editmode = EDIT_PATTERN;
    break;

    case VK_DOWN:
    case VK_RETURN:
    enpos++;
    if (enpos > 2) enpos = 0;
    break;

    case VK_UP:
    enpos--;
    if (enpos < 0) enpos = 2;
    break;
  }
  switch(enpos)
  {
    case 0:
    editstring(songname, MAX_STR);
    break;

    case 1:
    editstring(authorname, MAX_STR);
    break;

    case 2:
    editstring(copyrightname, MAX_STR);
    break;
  }
}

void patterncommands(void)
{
  int c;

  if (followplay)
  {
    if (virtualkey == VK_TAB) editmode++;
    return;
  }

  switch(key)
  {
    case '<':
    case '(':
    case '[':
    if (epnum[epchn] > 0)
    {
      epnum[epchn]--;
      if (eppos > pattlen[epnum[epchn]]) eppos = pattlen[epnum[epchn]];
    }
    break;

    case '>':
    case ')':
    case ']':
    if (epnum[epchn] < MAX_PATT - 1)
    {
      epnum[epchn]++;
      if (eppos > pattlen[epnum[epchn]]) eppos = pattlen[epnum[epchn]];
    }
    break;
  }
  {
    int newnote = -1;
    for (c = 0; c < sizeof(notekeytbl1); c++)
    {
      if ((tolower(key) == notekeytbl1[c]) && (!epcolumn))
      {
        newnote = c + epoctave * 12;
      }
    }
    for (c = 0; c < sizeof(notekeytbl2); c++)
    {
      if ((tolower(key) == notekeytbl2[c]) && (!epcolumn))
      {
        newnote = c + (epoctave+1) * 12;
      }
    }
    if (newnote >= KEYOFF) newnote = -1;
    if (virtualkey == 0x08) newnote = REST;
    if (virtualkey == 0x14) newnote = KEYOFF;
    if (virtualkey == VK_RETURN) newnote = KEYOFF;
    if (newnote >= 0)
    {
      if ((recordmode) && (eppos < pattlen[epnum[epchn]]))
      {
        pattern[epnum[epchn]][eppos * 3] = newnote;
        if (newnote < KEYOFF)
        {
          pattern[epnum[epchn]][eppos * 3 + 1] &= 0x07;
          pattern[epnum[epchn]][eppos * 3 + 1] |= einum << 3;
        }
        if (newnote == REST)
        {
          pattern[epnum[epchn]][eppos * 3 + 1] &= 0x07;
        }
      }
      if (recordmode)
      {
        eppos++;
        if (eppos > pattlen[epnum[epchn]])
        {
          eppos = 0;
        }
      }
      if (songinit == 0x80)
      {
        sidreg[0x4 + epchn * 7] = chn[epchn].wave & 0xfe; /* Hard restart */
        sidreg[0x5 + epchn * 7] = 0;
        sidreg[0x6 + epchn * 7] = 0;
        if (newnote < KEYOFF)
        {
          chn[epchn].newnote = 0xfe; /* Trigger note delayed by two updates */
          chn[epchn].instrnum = einum;
          chn[epchn].note = newnote;
        }
      }
    }
  }
  switch(virtualkey)
  {
    case VK_TAB:
    editmode++;
    break;

    case VK_F4:
    chn[epchn].mute ^= 1;
    break;

    case VK_F7:
    for (c = 0; c < pattlen[epnum[epchn]]; c++)
    {
      if ((pattern[epnum[epchn]][c*3] < KEYOFF) &&
          (pattern[epnum[epchn]][c*3] > 0))
        pattern[epnum[epchn]][c*3]--;
    }
    break;

    case VK_F8:
    for (c = 0; c < pattlen[epnum[epchn]]; c++)
    {
      if (pattern[epnum[epchn]][c*3] < KEYOFF-1)
        pattern[epnum[epchn]][c*3]++;
    }
    break;

    case VK_F5:
    memcpy(patterncopybuffer, pattern[epnum[epchn]], MAX_PATTROWS*3+3);
    break;

    case VK_F6:
    memcpy(pattern[epnum[epchn]], patterncopybuffer, MAX_PATTROWS*3+3);
    if (eppos > pattlen[epnum[epchn]]) eppos = pattlen[epnum[epchn]];
    countthispattern();
    break;

    case VK_DELETE:
    if ((pattlen[epnum[epchn]] - eppos)*3-3 >= 0)
    {
      memmove(&pattern[epnum[epchn]][eppos*3],
        &pattern[epnum[epchn]][eppos*3+3],
        (pattlen[epnum[epchn]] - eppos)*3-3);
      pattern[epnum[epchn]][pattlen[epnum[epchn]]*3-3] = REST;
      pattern[epnum[epchn]][pattlen[epnum[epchn]]*3-2] = 0x00;
      pattern[epnum[epchn]][pattlen[epnum[epchn]]*3-1] = 0x00;
    }
    else
    {
      if (eppos == pattlen[epnum[epchn]])
      {
        if (pattlen[epnum[epchn]] > 1)
        {
          pattern[epnum[epchn]][pattlen[epnum[epchn]]*3-3] = ENDPATT;
          pattern[epnum[epchn]][pattlen[epnum[epchn]]*3-2] = 0x00;
          pattern[epnum[epchn]][pattlen[epnum[epchn]]*3-1] = 0x00;
          countthispattern();
          eppos = pattlen[epnum[epchn]];
        }
      }
    }
    break;

    case VK_INSERT:
    if ((pattlen[epnum[epchn]] - eppos)*3-3 >= 0)
    {
      memmove(&pattern[epnum[epchn]][eppos*3+3],
        &pattern[epnum[epchn]][eppos*3],
        (pattlen[epnum[epchn]] - eppos)*3-3);
      pattern[epnum[epchn]][eppos*3] = REST;
      pattern[epnum[epchn]][eppos*3+1] = 0x00;
      pattern[epnum[epchn]][eppos*3+2] = 0x00;
    }
    else
    {
      if (eppos == pattlen[epnum[epchn]])
      {
        if (pattlen[epnum[epchn]] < MAX_PATTROWS)
        {
          pattern[epnum[epchn]][eppos*3] = REST;
          pattern[epnum[epchn]][eppos*3+1] = 0x00;
          pattern[epnum[epchn]][eppos*3+2] = 0x00;
          pattern[epnum[epchn]][eppos*3+3] = ENDPATT;
          pattern[epnum[epchn]][eppos*3+4] = 0x00;
          pattern[epnum[epchn]][eppos*3+5] = 0x00;
          countthispattern();
          eppos = pattlen[epnum[epchn]];
        }
      }
    }
    break;

    case VK_SPACE:
    recordmode ^= 1;
    break;

    case VK_RIGHT:
    epcolumn++;
    if (epcolumn >= 6)
    {
      epcolumn = 0;
      epchn++;
      if (epchn >= MAX_CHN) epchn = 0;
      if (eppos > pattlen[epnum[epchn]]) eppos = pattlen[epnum[epchn]];
    }
    break;

    case VK_LEFT:
    epcolumn--;
    if (epcolumn < 0)
    {
      epcolumn = 5;
      epchn--;
      if (epchn < 0) epchn = MAX_CHN - 1;
      if (eppos > pattlen[epnum[epchn]]) eppos = pattlen[epnum[epchn]];
    }
    break;

    case VK_PAGEDOWN:
    for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
    {
      eppos++;
      if (eppos > pattlen[epnum[epchn]])
      {
        eppos = 0;
      }
    }
    break;

    case VK_DOWN:
    eppos++;
    if (eppos > pattlen[epnum[epchn]])
    {
      eppos = 0;
    }
    break;

    case VK_HOME:
    eppos = 0;
    break;

    case VK_END:
    eppos = pattlen[epnum[epchn]];
    break;

    case VK_PAGEUP:
    for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
    {
      eppos--;
      if (eppos < 0)
      {
        eppos = pattlen[epnum[epchn]];
      }
    }
    break;

    case VK_UP:
    eppos--;
    if (eppos < 0)
    {
      eppos = pattlen[epnum[epchn]];
    }
    break;
  }
  if ((hexnybble >= 0) && (epcolumn))
  {
    if (eppos < pattlen[epnum[epchn]])
    {
      switch(epcolumn)
      {
        case 1:
        if (hexnybble < 2)
        {
          pattern[epnum[epchn]][eppos * 3 + 1] &= 0x7f;
          pattern[epnum[epchn]][eppos * 3 + 1] |= hexnybble << 7;
        }
        break;

        case 2:
        pattern[epnum[epchn]][eppos * 3 + 1] &= 0x87;
        pattern[epnum[epchn]][eppos * 3 + 1] |= hexnybble << 3;
        break;

        case 3:
        if (hexnybble < 8)
        {
          pattern[epnum[epchn]][eppos * 3 + 1] &= 0xf8;
          pattern[epnum[epchn]][eppos * 3 + 1] |= hexnybble;
          /* Check for illegal command data */
          if ((pattern[epnum[epchn]][eppos * 3 + 1] & 0x7) == CMD_SETTEMPO)
          {
            int tempo = pattern[epnum[epchn]][eppos * 3 + 2] & 0x7f;
            if (tempo < 3) tempo = 3;
            pattern[epnum[epchn]][eppos * 3 + 2] &= 0x80;
            pattern[epnum[epchn]][eppos * 3 + 2] |= tempo;
          }
        }
        break;

        case 4:
        pattern[epnum[epchn]][eppos * 3 + 2] &= 0x0f;
        pattern[epnum[epchn]][eppos * 3 + 2] |= hexnybble << 4;
        /* Check for illegal command data */
        if ((pattern[epnum[epchn]][eppos * 3 + 1] & 0x7) == CMD_SETTEMPO)
        {
          int tempo = pattern[epnum[epchn]][eppos * 3 + 2] & 0x7f;
          if (tempo < 3) tempo = 3;
          pattern[epnum[epchn]][eppos * 3 + 2] &= 0x80;
          pattern[epnum[epchn]][eppos * 3 + 2] |= tempo;
        }
        break;

        case 5:
        pattern[epnum[epchn]][eppos * 3 + 2] &= 0xf0;
        pattern[epnum[epchn]][eppos * 3 + 2] |= hexnybble;
        /* Check for illegal command data */
        if ((pattern[epnum[epchn]][eppos * 3 + 1] & 0x7) == CMD_SETTEMPO)
        {
          int tempo = pattern[epnum[epchn]][eppos * 3 + 2] & 0x7f;
          if (tempo < 3) tempo = 3;
          pattern[epnum[epchn]][eppos * 3 + 2] &= 0x80;
          pattern[epnum[epchn]][eppos * 3 + 2] |= tempo;
        }
        break;
      }
    }
    eppos++;
    if (eppos > pattlen[epnum[epchn]])
    {
      eppos = 0;
    }
  }
  epview = eppos - VISIBLEPATTROWS / 2;
}

void printmainscreen(void)
{
  clearscreen();
  printblankc(0, 0, 15+16, 80);
  printtext(0, 0, 15+16, programname);
  printtext(70, 0, 15+16, "F12 = HELP");
  printpatterns();
  fliptoscreen();
}

void editstring(char *buffer, int maxlength)
{
  int len = strlen(buffer);

  if (key)
  {
    if ((key >= 32) && (key < 256))
    {
      if (len < maxlength-1)
      {
        buffer[len] = key;
        buffer[len+1] = 0;
      }
    }
    if ((key == 8) && (len > 0))
    {
      buffer[len-1] = 0;
    }
  }
}

void getkey(void)
{
  virtualkey = 0;
  key = 0;
  win_asciikey = 0;
  win_virtualkey = 0;
  cursorflashdelay += win_getspeed(50);
  virtualkey = kbd_getvirtualkey();
  if (!virtualkey) return;
  win_checkmessages();
  key = kbd_getascii();
}


void waitkey(void)
{
  for (;;)
  {
    display_update();
    getkey();
    if (virtualkey) break;
    if (win_quitted) break;
  }
}

void waitkey2(void)
{
  for (;;)
  {
    getkey();
    if (virtualkey) break;
    if (win_quitted) break;
  }
}

void handle_int(void)
{
  exit(0); /* Atexit functions will be called! */
}

void playroutine(void)
{
  INSTR *iptr = &instr[0];
  CHN *cptr = &chn[0];
  int c;

  if (songinit == 0x03) followplay = 0;

  if ((songinit > 0) && (songinit < 0x80))
  {
    timeframe = 0;
    timemin = 0;
    timesec = 0;
    playsongnum = esnum;
    if (songinit == 0x02)
    {
      if ((espos[0] >= songlen[playsongnum][0]) ||
         (espos[1] >= songlen[playsongnum][1]) ||
         (espos[2] >= songlen[playsongnum][2]))
         songinit = 0x01;
    }


    for (c = 0; c < MAX_CHN; c++)
    {
      if (songinit == 0x02) cptr->songptr = espos[c];
      else cptr->songptr = 0;
      cptr->wavetable = 0;
      cptr->tick = 5;
      cptr->newnote = 5;
      if (songinit == 0x01) cptr->tempo = 5;
      sidreg[0x4 + c*7] = cptr->wave = 0x08; /* Testbit */
      cptr->pattptr = ENDPATT;
      cptr++;
    }
    sidreg[0x15] = 0;
    if (songinit == 0x01)
    {
      filterctrl = 0;
      filtercutoffadd = 0;
    }
    if (songinit != 0x03) songinit = 0;
    else songinit = 0x80;
    if ((!songlen[playsongnum][0]) ||
       (!songlen[playsongnum][1]) ||
       (!songlen[playsongnum][2])) songinit = 0x80; /* Zero length song */
  }
  else
  {
    filtercutoff += filtercutoffadd;
    sidreg[0x16] = filtercutoff;
    sidreg[0x17] = filterctrl;
    sidreg[0x18] = volume | filtertype;

    for (c = 0; c < MAX_CHN; c++)
    {
      iptr = &instr[cptr->instrnum];
      /* If playstate is negative, just do effects */
      if (songinit >= 0x80)
      {
        cptr->command = 0;
        cptr->cmddata = 0;
        if (cptr->newnote >= 0x80)
        {
          cptr->wavetable = 0;
          cptr->pulsedir = 0;
          cptr->newnote++;
        }
        goto EFFECTS;
      }

      /* Decrease tick */
      if (cptr->tick >= 0x80) cptr->tick = cptr->tempo;
      cptr->tick--;
      if (cptr->tick >= 0x80) goto NEWNOTES;

      /* Check for end of pattern */
      if (cptr->pattptr != ENDPATT) goto EFFECTS;
      cptr->pattptr = 0;
      if (songorder[playsongnum][c][cptr->songptr] < ENDSONG)
      {
        cptr->pattnum = songorder[playsongnum][c][cptr->songptr];
        cptr->songptr++;
      }
      else
      {
        if (songorder[playsongnum][c][cptr->songptr] == LOOPSONG)
        {
          cptr->songptr = 0;
          cptr->pattnum = songorder[playsongnum][c][cptr->songptr];
          cptr->songptr++;
        }
      }
      if (!cptr->newnote) goto EFFECTS;
      goto PULSEDONE;

      /* New notes processing */
      NEWNOTES:
      {
        unsigned char newinstrnum;
        unsigned char newnote;

        newnote = pattern[cptr->pattnum][cptr->pattptr];
        newinstrnum = pattern[cptr->pattnum][cptr->pattptr+1] >> 3;
        if (newinstrnum) cptr->instrnum = newinstrnum;
        cptr->command = pattern[cptr->pattnum][cptr->pattptr+1] & 0x7;
        cptr->cmddata = pattern[cptr->pattnum][cptr->pattptr+2];
        cptr->pattptr += 3;
        if (pattern[cptr->pattnum][cptr->pattptr] == ENDPATT)
          cptr->pattptr = ENDPATT;

        if (newnote < REST)
        {
          if (newnote == KEYOFF)
            sidreg[0x4 + 7*c] = cptr->wave & 0xfe;
          else
          {
            cptr->note = newnote;
            if (cptr->command != CMD_TONEPORTA)
            {
              cptr->newnote = 0;
              sidreg[0x4 + 7*c] = cptr->wave & 0xfe;
              sidreg[0x5 + 7*c] = 0; /* Hard restart */
              sidreg[0x6 + 7*c] = 0;
            }
          }
        }
      }
      /* Tick 0 effects */
      switch (cptr->command)
      {
        case CMD_SETCUTOFFADD:
        filtercutoffadd = cptr->cmddata;
        break;

        case CMD_TONEPORTA:
        cptr->vibdir = 0xfe;
        break;

        case CMD_SETFILTER:
        filterctrl = cptr->cmddata;
        break;

        case CMD_SETSUSTAIN:
        sidreg[0x6 + 7*c] = cptr->cmddata;
        break;

        case CMD_SETTEMPO:
        /* Refuse to use tempo < 3 because it wouldn't work correctly */
        if ((cptr->cmddata & 0x7f) >= 3)
        {
          if (cptr->cmddata >= 0x80)
            cptr->tempo = (cptr->cmddata & 0x7f) - 1;
          else
          {
            chn[0].tempo = cptr->cmddata - 1;
            chn[1].tempo = cptr->cmddata - 1;
            chn[2].tempo = cptr->cmddata - 1;
          }
        }
        break;
      }
      goto NEXTCHN;

      /* Continuous effects */
      EFFECTS:
      if (!cptr->newnote)
      {
        cptr->newnote = 0x3f;
        cptr->pulsedir = 0xfe;
        cptr->arpcount = 0xfe;
        cptr->vibdir = 0xfe;
        cptr->wavetable = 1;
        cptr->waveinstr = cptr->instrnum;
        cptr->pulse = iptr->pulse;
        sidreg[0x2 + 7*c] = cptr->pulse;
        sidreg[0x3 + 7*c] = cptr->pulse;
        if (iptr->filtertype)
        {
          filtercutoff = iptr->filtertype & 0xf0;
          filtertype = iptr->filtertype << 4;
        }
        goto NEXTCHN;
      }
      PULSEMOD:
      {
        int newpulse;

        if (cptr->pulsedir >= 0x80)
        {
          sidreg[0x5 + 7*c] = iptr->ad;
          sidreg[0x6 + 7*c] = iptr->sr;
          cptr->pulsedir = 0;
        }
        else
        {
          if (!cptr->pulsedir)
          {
            newpulse = cptr->pulse;
            newpulse += iptr->pulseadd;
            if (newpulse > 0xff) newpulse++;
            cptr->pulse = newpulse & 0xff;
            if ((cptr->pulse & 0xf) >= iptr->pulselimithigh)
              cptr->pulsedir = 0x01;
            sidreg[0x2 + 7*c] = cptr->pulse;
            sidreg[0x3 + 7*c] = cptr->pulse;
          }
          else
          {
            newpulse = cptr->pulse;
            newpulse -= iptr->pulseadd;
            if (newpulse < 0) newpulse--;
            cptr->pulse = newpulse & 0xff;
            if ((cptr->pulse & 0xf) < iptr->pulselimitlow)
              cptr->pulsedir = 0x00;
            sidreg[0x2 + 7*c] = cptr->pulse;
            sidreg[0x3 + 7*c] = cptr->pulse;
          }
        }
      }
      PULSEDONE:
      if (cptr->wavetable) goto WAVETABLE;
      if (!cptr->command) goto ARPEGGIO;
      switch(cptr->command)
      {
        case CMD_PORTAMENTO:
        {
          unsigned char speed = cptr->cmddata << 1;
          if (cptr->cmddata < 0x80) freqadd(c, cptr, speed);
          else freqsub(c, cptr, speed);
        }
        break;

        case CMD_VIBRATO:
        {
          unsigned char speed = swapnybbles(cptr->cmddata);

          if ((cptr->vibdir < 0x80) && (cptr->vibdir >= ((cptr->cmddata >> 4) & 0x0e)))
          {
            cptr->vibdir ^= 0xff;
          }
          else
          {
            cptr->vibdir += 0x02;
          }
          if (cptr->vibdir & 0x01) freqsub(c, cptr, speed);
          else freqadd(c, cptr, speed);
        }
        break;

        case CMD_TONEPORTA:
        {
          unsigned short targetfreq = freqtbllo[cptr->note] | (freqtblhi[cptr->note] << 8);
          unsigned short currentfreq = cptr->freqlo | (cptr->freqhi << 8);
          unsigned char speed = cptr->cmddata << 1;

          if (cptr->cmddata < 0x80)
          {
            if (currentfreq < targetfreq) freqadd(c, cptr, speed);
            else
            {
              sidreg[0x0 + 7*c] = cptr->freqlo = targetfreq & 0xff;
              sidreg[0x1 + 7*c] = cptr->freqhi = targetfreq >> 8;
            }
          }
          else
          {
            if (currentfreq > targetfreq) freqsub(c, cptr, speed);
            else
            {
              sidreg[0x0 + 7*c] = cptr->freqlo = targetfreq & 0xff;
              sidreg[0x1 + 7*c] = cptr->freqhi = targetfreq >> 8;
            }
          }
        }
        break;
      }
      goto NEXTCHN;

      WAVETABLE:
      {
        unsigned char note;

        if (wavetable[cptr->waveinstr][cptr->wavetable - 1])
        {
          if (!cptr->mute) sidreg[0x4 + 7*c] = cptr->wave = wavetable[cptr->waveinstr][cptr->wavetable - 1];
          else sidreg[0x4 + 7*c] = cptr->wave = 0x08;
        }
        note = wavetable[cptr->waveinstr][cptr->wavetable];
        if (note < 0x80)
          note += cptr->note;
        note &= 0x7f;

        switch (wavetable[cptr->waveinstr][cptr->wavetable+1])
        {
          case 0xff:
          if (wavetable[cptr->waveinstr][cptr->wavetable+2])
          {
            cptr->waveinstr = cptr->instrnum;
            cptr->wavetable = (wavetable[cptr->waveinstr][cptr->wavetable+2] - 1) * 2 + 1;
          }
          else
          {
            cptr->wavetable = 0;
          }
          break;

          default:
          cptr->wavetable += 2;
          break;
        }
        sidreg[0x0 + 7*c] = cptr->freqlo = freqtbllo[note];
        sidreg[0x1 + 7*c] = cptr->freqhi = freqtblhi[note];
      }
      goto NEXTCHN;

      ARPEGGIO:
      if (cptr->cmddata)
      {
        unsigned char note = cptr->note;

        /* Slow arpeggio */
        if (cptr->cmddata >= 0x80)
        {
          if (cptr->tick & 1) goto SLOWARPDELAY;
        }

        if (!cptr->arpcount)
        {
          cptr->arpcount = 0x01;
          goto ARPDONE;
        }
        if (cptr->arpcount < 0x80)
        {
          note += (cptr->cmddata & 0x70) >> 4;
          cptr->arpcount = 0xff;
          goto ARPDONE;
        }
        note += (cptr->cmddata & 0xf);
        cptr->arpcount = 0;

        ARPDONE:
        note &= 0x7f;
        sidreg[0x0 + 7*c] = cptr->freqlo = freqtbllo[note];
        sidreg[0x1 + 7*c] = cptr->freqhi = freqtblhi[note];
        SLOWARPDELAY:
      }
      NEXTCHN:
      if (cptr->mute) sidreg[0x4 + 7*c] = cptr->wave = 0x08;

      cptr++;
    }
  }
  if (songinit != 0x80)
  {
    timeframe++;
    if (timeframe == 50)
    {
      timeframe = 0;
      timesec++;
    }
    if (timesec == 60)
    {
      timesec = 0;
      timemin++;
      timemin %= 60;
    }
  }
}

void freqadd(int c, CHN *cptr, unsigned char speed)
{
  int newlowbyte = cptr->freqlo;
  newlowbyte += speed;

  sidreg[0x0 + 7*c] = cptr->freqlo;
  sidreg[0x1 + 7*c] = cptr->freqhi;
  if (newlowbyte > 0xff) cptr->freqhi++;
  cptr->freqlo = newlowbyte & 0xff;
}

void freqsub(int c, CHN *cptr, unsigned char speed)
{
  int newlowbyte = cptr->freqlo;
  newlowbyte -= speed;

  sidreg[0x0 + 7*c] = cptr->freqlo;
  sidreg[0x1 + 7*c] = cptr->freqhi;
  if (newlowbyte < 0) cptr->freqhi--;
  cptr->freqlo = newlowbyte & 0xff;
}

unsigned char swapnybbles(unsigned char n)
{
  unsigned char highnybble = n >> 4;
  unsigned char lownybble = n & 0xf;

  return (lownybble << 4) | highnybble;
}

void clearsong(int cs, int cp, int ci, int cn)
{
  int c;

  songinit = 0x3;

  for (c = 0; c < MAX_CHN; c++)
  {
    int d;
    chn[c].mute = 0;
    chn[c].tempo = 5;
    if (cs)
    {
      for (d = 0; d < MAX_SONGS; d++)
      {
        memset(&songorder[d][c][0], 0, MAX_SONGLEN);
        if (!d)
        {
          songorder[d][c][0] = c;
          songorder[d][c][1] = LOOPSONG;
        }
        else
        {
          songorder[d][c][0] = ENDSONG;
        }
      }
      epnum[c] = songorder[0][c][0];
      espos[c] = 0;
    }
  }
  if (cs)
  {
    esview = 0;
    eseditpos = 0;
    escolumn = 0;
    eschn = 0;
    esnum = 0;
    eppos = 0;
    epview = - VISIBLEPATTROWS / 2;
    epcolumn = 0;
    epchn = 0;
  }
  if (cn)
  {
    memset(songname, 0, sizeof songname);
    memset(authorname, 0, sizeof authorname);
    memset(copyrightname, 0, sizeof copyrightname);
    enpos = 0;
  }

  if (cp)
  {
    for (c = 0; c < MAX_PATT; c++)
    {
      int d;
      memset(&pattern[c][0], 0, MAX_PATTROWS*3);
      for (d = 0; d < 64; d++) pattern[c][d*3] = REST;
      for (d = 64; d < MAX_PATTROWS; d++) pattern[c][d*3] = ENDPATT;
    }
    memcpy(patterncopybuffer, pattern[0], MAX_PATTROWS*3);
  }
  if (ci)
  {
    for (c = 0; c < MAX_INSTR; c++)
    {
      memset(&instr[c], 0, sizeof(INSTR));
      memset(&wavetable[c][0], 0, MAX_WAVELEN*2);
      wavetable[c][2] = 0xff;
      wavetable[c][MAX_WAVELEN*2-2] = 0xff;
    }
    memset(&instrcopybuffer, 0, sizeof(INSTR));
    memset(wavecopybuffer, 0, MAX_WAVELEN*2);
    eipos = 0;
    eicolumn = 0;
    einum = 1;
    ewpos = 0;
  }
  countpatternlengths();
  playsongnum = 0;
}

void countpatternlengths(void)
{
  int c, d, e;

  highestusedpattern = 0;
  for (c = 0; c < MAX_PATT; c++)
  {
    for (d = 0; d <= MAX_PATTROWS; d++)
    {
      if (pattern[c][d*3] == ENDPATT) break;
    }
    pattlen[c] = d;
  }
  for (e = 0; e < MAX_SONGS; e++)
  {
    for (c = 0; c < MAX_CHN; c++)
    {
      for (d = 0; d < MAX_SONGLEN; d++)
      {
        if (songorder[e][c][d] >= ENDSONG) break;
        if (songorder[e][c][d] > highestusedpattern)
          highestusedpattern = songorder[e][c][d];
      }
      songlen[e][c] = d;
    }
  }
}

void countthispattern(void)
{
  int c, d, e;

  c = epnum[epchn];
  for (d = 0; d <= MAX_PATTROWS; d++)
  {
    if (pattern[c][d*3] == ENDPATT) break;
  }
  pattlen[c] = d;

  e = esnum;
  c = eschn;
  for (d = 0; d < MAX_SONGLEN; d++)
  {
    if (songorder[e][c][d] >= ENDSONG) break;
    if (songorder[e][c][d] > highestusedpattern)
      highestusedpattern = songorder[e][c][d];
  }
  songlen[e][c] = d;
}


void printpatterns(void)
{
  int c, d, color;
  int cc = cursorcolortable[cursorflash];

  for (c = 0; c < MAX_CHN; c++)
  {
    if (!followplay)
    {
      sprintf(textbuffer, "CHN%d PATT%02X", c+1, epnum[c]);
      printtext(1+c*13, 2, 15, textbuffer);
      for (d = 0; d < VISIBLEPATTROWS; d++)
      {
        int rownum = epview + d;
        color = 7;
        if (chn[c].mute) color = 8;
        if (rownum == eppos) color = 10;
        if ((rownum < 0) || (rownum > pattlen[epnum[c]]))
        {
          sprintf(textbuffer, "           ");
        }
        else
        {
          if (pattern[epnum[c]][rownum*3] == ENDPATT)
          {
            sprintf(textbuffer, "%02d PATT.END",
              rownum,
              pattern[epnum[c]][rownum*3+1] >> 3,
              pattern[epnum[c]][rownum*3+1] & 0x7,
              pattern[epnum[c]][rownum*3+2]);
          }
          else
          {
            sprintf(textbuffer, "%02d %s%02X%01X%02X",
              rownum,
              notename[pattern[epnum[c]][rownum*3]],
              pattern[epnum[c]][rownum*3+1] >> 3,
              pattern[epnum[c]][rownum*3+1] & 0x7,
              pattern[epnum[c]][rownum*3+2]);
          }
        }
        printtext(1+c*13, 3+d, color, textbuffer);
        if ((color == 10) && (editmode == EDIT_PATTERN) && (epchn == c))
        {
          switch(epcolumn)
          {
            case 0:
            printbg(1+c*13+3, 3+d, cc, 3);
            break;

            default:
            printbg(1+c*13+5+epcolumn, 3+d, cc, 1);
            break;
          }
        }
      }
    }
    else
    {
      sprintf(textbuffer, "CHN%d PATT%02X", c+1, chn[c].pattnum);
      printtext(1+c*13, 2, 15, textbuffer);
      for (d = 0; d < VISIBLEPATTROWS; d++)
      {
        int pos = chn[c].pattptr;
        int view, rownum;

        if (pos == ENDPATT) pos = pattlen[chn[c].pattnum];
          else pos /= 3;
        view = pos - VISIBLEPATTROWS / 2;
        epnum[c] = chn[c].pattnum;
        if (epchn == c)
        {
          eppos = pos;
          epview = view;
        }

        rownum = view + d;
        color = 7;
        if (chn[c].mute) color = 8;
        if (rownum == pos) color = 10;
        if ((rownum < 0) || (rownum > pattlen[chn[c].pattnum]))
        {
          sprintf(textbuffer, "           ");
        }
        else
        {
          if (pattern[chn[c].pattnum][rownum*3] == ENDPATT)
          {
            sprintf(textbuffer, "%02d PATT.END",
              rownum,
              pattern[chn[c].pattnum][rownum*3+1] >> 3,
              pattern[chn[c].pattnum][rownum*3+1] & 0x7,
              pattern[chn[c].pattnum][rownum*3+2]);
          }
          else
          {
            sprintf(textbuffer, "%02d %s%02X%01X%02X",
              rownum,
              notename[pattern[chn[c].pattnum][rownum*3]],
              pattern[chn[c].pattnum][rownum*3+1] >> 3,
              pattern[chn[c].pattnum][rownum*3+1] & 0x7,
              pattern[chn[c].pattnum][rownum*3+2]);
          }
        }
        printtext(1+c*13, 3+d, color, textbuffer);
      }
    }
  }
  sprintf(textbuffer, "CHN ORDERLIST (SUBTUNE %02X, POS %03d)", esnum+1, eseditpos);
  printtext(40, 2, 15, textbuffer);
  for (c = 0; c < MAX_CHN; c++)
  {
    sprintf(textbuffer, " %d ", c+1);
    printtext(40, 4+c, 15, textbuffer);
    for (d = 0; d < VISIBLEORDERLIST; d++)
    {
      int rownum = esview + d;
      color = 7;
      if (rownum == espos[c]) color = 10;
      if ((rownum < 0) || (rownum > songlen[esnum][c]) || (rownum > MAX_SONGLEN))
      {
        sprintf(textbuffer, "   ");
      }
      else
      {
        if (songorder[esnum][c][rownum] < ENDSONG)
        {
          sprintf(textbuffer, "%02X ", songorder[esnum][c][rownum]);
        }
        if (songorder[esnum][c][rownum] == ENDSONG)
        {
          sprintf(textbuffer, "STP", songorder[esnum][c][rownum]);
        }
        if (songorder[esnum][c][rownum] == LOOPSONG)
        {
          sprintf(textbuffer, "RST", songorder[esnum][c][rownum]);
        }
      }
      printtext(44+d*3, 4+c, color, textbuffer);
      if ((rownum == eseditpos) && (editmode == EDIT_ORDERLIST) && (eschn == c))
      {
        printbg(44+d*3+escolumn, 4+c, cc, 1);
      }
    }
  }
  sprintf(textbuffer, "INSTRUMENT NUM. %02X  %-16s", einum, instr[einum].name);
  printtext(40, 9, 15, textbuffer);

  sprintf(textbuffer, "Attack/Decay    %02X", instr[einum].ad);
  if (eipos == 0) color = 10; else color = 7;
  printtext(40, 11, color, textbuffer);

  sprintf(textbuffer, "Sustain/Release %02X", instr[einum].sr);
  if (eipos == 1) color = 10; else color = 7;
  printtext(40, 12, color, textbuffer);

  sprintf(textbuffer, "Pulse Width     %02X", swapnybbles(instr[einum].pulse));
  if (eipos == 2) color = 10; else color = 7;
  printtext(40, 13, color, textbuffer);

  sprintf(textbuffer, "Pulse Speed     %02X", swapnybbles(instr[einum].pulseadd));
  if (eipos == 3) color = 10; else color = 7;
  printtext(40, 14, color, textbuffer);

  sprintf(textbuffer, "Pulse Limit Min %02X", swapnybbles(instr[einum].pulselimitlow));
  if (eipos == 4) color = 10; else color = 7;
  printtext(40, 15, color, textbuffer);

  sprintf(textbuffer, "Pulse Limit Max %02X", swapnybbles(instr[einum].pulselimithigh));
  if (eipos == 5) color = 10; else color = 7;
  printtext(40, 16, color, textbuffer);

  sprintf(textbuffer, "Filt. Freq/Type %02X", instr[einum].filtertype);
  if (eipos == 6) color = 10; else color = 7;
  printtext(40, 17, color, textbuffer);

  printtext(60, 11, 15, "W       N POS");
  sprintf(textbuffer,"A       O %02X", ewpos+1);
  printtext(60, 12,  15, textbuffer);
  printtext(60, 13, 15, "V       T");
  printtext(60, 14, 15, "E       E");
  printtext(60, 15, 15, "T       T");
  printtext(60, 16, 15, "B       B");
  printtext(60, 17, 15, "L       L");
  for (d = 0; d < VISIBLEWAVEROWS; d++)
  {
    int rownum = ewview + d;
    color = 7;
    if (rownum == ewpos) color = 10;
    if ((rownum < 0) || (rownum >= MAX_WAVELEN))
    {
      sprintf(textbuffer, "     ");
    }
    else
    {
      sprintf(textbuffer, "%02X %02X",
      wavetable[einum][rownum*2],
      wavetable[einum][rownum*2+1]);
    }
    printtext(62, 11+d, color, textbuffer);
  }
  if (editmode == EDIT_INSTRUMENT)
  {
    if (eipos < 7)
    {
      switch (eicolumn)
      {
        case 0:
        case 1:
        printbg(56+eicolumn, 11+eipos, cc, 1);
        break;

        case 2:
        case 3:
        printbg(60+eicolumn, 11+ewpos-ewview, cc, 1);
        break;

        case 4:
        case 5:
        printbg(61+eicolumn, 11+ewpos-ewview, cc, 1);
        break;
      }
    }
    else
    {
      printbg(60+strlen(instr[einum].name), 9, cc, 1);
    }
  }

  if (enpos == 0) color = 10; else color = 7;
  printtext(40, 20, 15, "SONG NAME");
  sprintf(textbuffer, "%-32s", songname);
  printtext(40, 21, color, textbuffer);

  if (enpos == 1) color = 10; else color = 7;
  printtext(40, 22, 15, "AUTHOR");
  sprintf(textbuffer, "%-32s", authorname);
  printtext(40, 23, color, textbuffer);

  if (enpos == 2) color = 10; else color = 7;
  printtext(40, 24, 15, "COPYRIGHT");
  sprintf(textbuffer, "%-32s", copyrightname);
  printtext(40, 25, color, textbuffer);

  if (editmode == EDIT_NAMES)
  {
    switch(enpos)
    {
      case 0:
      printbg(40+strlen(songname), 21, cc, 1);
      break;

      case 1:
      printbg(40+strlen(authorname), 23, cc, 1);
      break;

      case 2:
      printbg(40+strlen(copyrightname), 25, cc, 1);
      break;
    }
  }
  sprintf(textbuffer, "OCTAVE %d", epoctave);
  printtext(0, 28, 15, textbuffer);
  if (recordmode) printtext(0, 29, 10, "EDITMODE");
  else printtext(0, 29, 10, "JAM MODE");
  if (!songinit) printtext(10, 28, 15, "PLAYING");
  else printtext(10, 28, 15, "STOPPED");
  sprintf(textbuffer, " %02d%c%02d ", timemin, timechar[timeframe / 25], timesec);
  printtext(10, 29, 10, textbuffer);

  printtext(60, 28, 15, " CHN1   CHN2   CHN3 ");
  for (c = 0; c < MAX_CHN; c++)
  {
    int chnpos = chn[c].songptr;
    int chnrow = chn[c].pattptr;
    chnpos--;
    if (chnpos < 0) chnpos = 0;
    if (chnrow == ENDPATT) chnrow = 0;
    chnrow /= 3;

    sprintf(textbuffer, "%03d/%02d",
      chnpos,chnrow);
    printtext(60+7*c, 29, 10, textbuffer);
  }
}

int fileselector(char *name, char *filter, char *title, int initialmode)
{
  WIN32_FIND_DATA findfiledata;
  HANDLE findhandle;
  DIRENTRY swaptemp;
  int color;
  int c, d;
  int files;
  int filepos;
  int fileview;
  int lowest;
  int filemode = initialmode;
  int exitfilesel;
  char drivestr[] = "A:";
  char driveexists[26];

  for (c = 0; c < VISIBLEFILES+6; c++)
  {
    printblank(40 - (MAX_FILENAME+10)/2, 2+c, MAX_FILENAME+10);
  }
  drawbox(40-(MAX_FILENAME+10)/2, 2, 15, MAX_FILENAME+10, VISIBLEFILES+6);
  printblankc(40 - (MAX_FILENAME+10)/2+1, 3, 15+16,MAX_FILENAME+8);
  printtext(40-(MAX_FILENAME+10)/2+1, 3, 15+16, title);

  /* Scan for all existing drives */
  for (c = 0; c < 26; c++)
  {
    drivestr[0] = 'A' + c;
    if (GetDriveType(drivestr) > 1) driveexists[c] = 1;
    else driveexists[c] = 0;
  }

  NEWPATH:
  files = 0;
  for (c = 0; c < 26; c++)
  {
    if (driveexists[c])
    {
      drivestr[0] = 'A' + c;
      strcpy(direntry[files].name, drivestr);
      direntry[files].attribute = 2;
      files++;
    }
  }
  /* Scan for subdirectories in the current path */
  findhandle = FindFirstFile("*.*", &findfiledata);
  if (findhandle != INVALID_HANDLE_VALUE)
  {
    for (;;)
    {
      int result;
      if ((files < MAX_FILES) && (strlen(findfiledata.cFileName) < MAX_FILENAME) && (findfiledata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
      {
        strcpy(direntry[files].name, findfiledata.cFileName);
        direntry[files].attribute = 1;
        files++;
      }
	    result = FindNextFile(findhandle, &findfiledata);
      if (!result) break;
    }
    FindClose(findhandle);
  }
  /* Scan for files in the current path */
  findhandle = FindFirstFile(filter, &findfiledata);
  if (findhandle != INVALID_HANDLE_VALUE)
  {
    for (;;)
    {
      int result;
      if ((files < MAX_FILES) && (strlen(findfiledata.cFileName) < MAX_FILENAME) && (!(findfiledata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)))
      {
        strcpy(direntry[files].name, findfiledata.cFileName);
        direntry[files].attribute = 0;
        files++;
      }
	    result = FindNextFile(findhandle, &findfiledata);
      if (!result) break;
    }
    FindClose(findhandle);
  }
  /* Sort the filelist in a most horrible fashion */
  for (c = 0; c < files; c++)
  {
    lowest = c;
    for (d = c+1; d < files; d++)
    {
      if (direntry[d].attribute < direntry[lowest].attribute)
      {
        lowest = d;
      }
      else
      {
        if (direntry[d].attribute == direntry[lowest].attribute)
        {
          if (strcmp(direntry[d].name, direntry[lowest].name) < 0)
          {
            lowest = d;
          }
        }
      }
    }
    if (lowest != c)
    {
      swaptemp = direntry[c];
      direntry[c] = direntry[lowest];
      direntry[lowest] = swaptemp;
    }
  }

  /* Search for the current filename */
  filepos = 0;
  for (c = 0; c < files; c++)
  {
    if ((!direntry[c].attribute) && (!strcmp(name, direntry[c].name)))
    {
      filepos = c;
    }
  }

  exitfilesel = -1;
  while (exitfilesel < 0)
  {
    int cc = cursorcolortable[cursorflash];

    if (cursorflashdelay >= 6)
    {
      cursorflashdelay %= 6;
      cursorflash++;
      cursorflash &= 3;
      fliptoscreen();
    }
    else
    {
      if (virtualkey) fliptoscreen();
    }
    getkey();

    switch(virtualkey)
    {
      case VK_ESCAPE:
      exitfilesel = 0;
      break;

      case VK_PAGEUP:
      for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
      {
        if ((!filemode) && (filepos > 0))
        {
          filepos--;
          if (!direntry[filepos].attribute) strcpy(name, direntry[filepos].name);
        }
      }
      break;

      case VK_UP:
      if ((!filemode) && (filepos > 0))
      {
        filepos--;
        if (!direntry[filepos].attribute) strcpy(name, direntry[filepos].name);
      }
      break;

      case VK_PAGEDOWN:
      for (scrrep = PAGEUPDNREPEAT; scrrep; scrrep--)
      {
        if ((!filemode) && (filepos < files-1))
        {
          filepos++;
          if (!direntry[filepos].attribute) strcpy(name, direntry[filepos].name);
        }
      }
      break;

      case VK_DOWN:
      if ((!filemode) && (filepos < files-1))
      {
        filepos++;
        if (!direntry[filepos].attribute) strcpy(name, direntry[filepos].name);
      }
      break;

      case VK_TAB:
      filemode++;
      if (filemode > 2) filemode = 0;
      break;

      case VK_RETURN:
      switch(filemode)
      {
        case 0:
        switch (direntry[filepos].attribute)
        {
          case 0:
          strcpy(name, direntry[filepos].name);
          exitfilesel = 1;
          break;

          case 1:
          SetCurrentDirectory(direntry[filepos].name);
          exitfilesel = 2;
          break;

          case 2:
          strcat(direntry[filepos].name, "\\");
          SetCurrentDirectory(direntry[filepos].name);
          exitfilesel = 2;
          break;
        }
        break;

        case 1:
        filemode = 0;
        exitfilesel = 2;
        break;

        case 2:
        exitfilesel = 1;
        break;
      }
      break;
    }
    if (filemode == 1)
    {
      editstring(filter, MAX_FILENAME);
    }
    if (filemode == 2)
    {
      editstring(name, MAX_FILENAME);
    }
    fileview = - VISIBLEFILES / 2 + filepos;


    for (c = 0; c < VISIBLEFILES; c++)
    {
      if ((fileview + c >= 0) && (fileview + c < files))
      {
        switch (direntry[fileview+c].attribute)
        {
          case 0:
          sprintf(textbuffer, "%-40s        ", direntry[fileview+c].name);
          break;

          case 1:
          sprintf(textbuffer, "%-40s   <DIR>", direntry[fileview+c].name);
          break;

          case 2:
          sprintf(textbuffer, "%-40s   <DRV>", direntry[fileview+c].name);
          break;
        }
      }
      else
      {
        sprintf(textbuffer, "                                                ");
      }
      color = 7;
      if ((fileview + c) == filepos) color = 10;
      printtext(40-(MAX_FILENAME+10)/2+1, 4+c, color, textbuffer);
      if ((!filemode) && ((fileview + c) == filepos)) printbg(40-(MAX_FILENAME+10)/2+1, 4+c, cc, 48);
    }

    printtext(40-(MAX_FILENAME+10)/2+1, 5+VISIBLEFILES, 15, "FILTER: ");
    sprintf(textbuffer, "%-40s", filter);
    color = 7;
    if (filemode == 1) color = 10;
    printtext(40-(MAX_FILENAME+10)/2+9, 5+VISIBLEFILES, color, textbuffer);
    if (filemode == 1) printbg(40-(MAX_FILENAME+10)/2+9+strlen(filter), 5+VISIBLEFILES, cc, 1);

    printtext(40-(MAX_FILENAME+10)/2+1, 6+VISIBLEFILES, 15, "NAME:   ");
    sprintf(textbuffer, "%-40s", name);
    color = 7;
    if (filemode == 2) color = 10;
    printtext(40-(MAX_FILENAME+10)/2+9, 6+VISIBLEFILES, color, textbuffer);
    if (filemode == 2) printbg(40-(MAX_FILENAME+10)/2+9+strlen(name), 6+VISIBLEFILES, cc, 1);

    if (win_quitted) exitfilesel = 0;
  }
  if (exitfilesel == 2) goto NEWPATH;

  printmainscreen();
  return exitfilesel;
}

void relocator(void)
{
  unsigned char packedsongname[MAX_FILENAME];
  unsigned char packedfilter[MAX_FILENAME];
  unsigned char wavetemp[MAX_WAVELEN * 2];

  int patterns = 0;
  int songs = 0;
  int pattsize = 0;
  int patttblsize = 0;
  int songsize = 0;
  int songtblsize = 0;
  int instrsize = 0;
  int wavetblsize = 0;
  int playersize = 0;
  int totalsize = 0;
  int playerhandle;
  int songhandle;
  int selectdone = 0;

  int c;

  unsigned instradr;
  unsigned wavetbladr;
  unsigned songtbladr;
  unsigned songadr;
  unsigned patttbladr;
  unsigned pattadr;

  unsigned char patttemp[256];

  unsigned char *playerwork = NULL;
  unsigned char *songwork = NULL;
  unsigned char *songtblwork = NULL;
  unsigned char *pattwork = NULL;
  unsigned char *patttblwork = NULL;
  unsigned char *instrwork = NULL;
  unsigned char *wavetblwork = NULL;

  songinit = 0x03;

  clearscreen();
  printblankc(0, 0, 15+16, 80);
  printtext(0, 0, 15+16, "GoatTracker PACKER/RELOCATOR");
  printtextc(15, 7, "Packing song...");
  fliptoscreen();

  /* Process song-orderlists */

  /* Calculate amount of songs with nonzero length */
  countpatternlengths();
  for (c = 0; c < MAX_SONGS; c++)
  {
    if ((songlen[c][0]) &&
        (songlen[c][1]) &&
        (songlen[c][2]))
    {
      songs++;
      songsize += songlen[c][0]+1;
      songsize += songlen[c][1]+1;
      songsize += songlen[c][2]+1;
    }
  }

  /* Allocate memory for songtable & song-orderlists */
  songtblsize = songs * 6;
  songtblwork = malloc(songtblsize);
  songwork = malloc(songsize);
  if ((!songtblwork) || (!songwork))
  {
    clearscreen();
    printtextc(15, 15, "OUT OF MEMORY IN PACKER/RELOCATOR!");
    fliptoscreen();
    waitkey2();
    goto PRCLEANUP;
  }

  /* Generate songorderlists & songtable */
  songsize = 0;
  for (c = 0; c < songs; c++)
  {
    if ((songlen[c][0]) &&
        (songlen[c][1]) &&
        (songlen[c][2]))
    {
      songtblwork[c*3] = songsize & 0xff;
      songtblwork[(c+songs)*3] = songsize >> 8;
      memcpy(&songwork[songsize], songorder[c][0], songlen[c][0]+1);
      songsize += songlen[c][0]+1;

      songtblwork[c*3 + 1] = songsize & 0xff;
      songtblwork[(c+songs)*3 + 1] = songsize >> 8;
      memcpy(&songwork[songsize], songorder[c][1], songlen[c][1]+1);
      songsize += songlen[c][1]+1;

      songtblwork[c*3 + 2] = songsize & 0xff;
      songtblwork[(c+songs)*3 + 2] = songsize >> 8;
      memcpy(&songwork[songsize], songorder[c][2], songlen[c][2]+1);
      songsize += songlen[c][2]+1;
    }
  }

  /* Calculate total size of patterns */
  for (c = 0; c <= highestusedpattern; c++)
  {
    pattsize += packpattern(patttemp, pattern[c], pattlen[c]);
    patterns++;
  }

  patttblsize = patterns * 2;
  patttblwork = malloc(patttblsize);
  pattwork = malloc(pattsize);
  if ((!patttblwork) || (!pattwork))
  {
    clearscreen();
    printtextc(15, 15, "OUT OF MEMORY IN PACKER/RELOCATOR!");
    fliptoscreen();
    waitkey2();
    goto PRCLEANUP;
  }

  /* This time pack the patterns for real */
  highestusedinstr = 0;
  pattsize = 0;
  for (c = 0; c < patterns; c++)
  {
    patttblwork[c] = pattsize & 0xff;
    patttblwork[c+patterns] = pattsize >> 8;
    pattsize += packpattern(&pattwork[pattsize], pattern[c], pattlen[c]);
  }
  highestusedinstr >>= 3;

  /* Then process instruments */
  if (!highestusedinstr) highestusedinstr = 1;

  instrsize = highestusedinstr * 8;
  instrwork = malloc(instrsize);
  wavetblwork = malloc(MAX_WAVELEN * highestusedinstr);
  if ((!instrwork) || (!wavetblwork))
  {
    clearscreen();
    printtextc(15, 15, "OUT OF MEMORY IN PACKER/RELOCATOR!");
    fliptoscreen();
    waitkey2();
    goto PRCLEANUP;
  }

  for (c = 1; c <= highestusedinstr; c++)
  {
    int d;
    int instrwavesize;

    instrwork[(c-1)*8+0] = instr[c].ad;
    instrwork[(c-1)*8+1] = instr[c].sr;
    instrwork[(c-1)*8+2] = instr[c].pulse;
    instrwork[(c-1)*8+3] = instr[c].pulseadd;
    instrwork[(c-1)*8+4] = instr[c].pulselimitlow;
    instrwork[(c-1)*8+5] = instr[c].pulselimithigh;
    instrwork[(c-1)*8+6] = instr[c].filtertype;

    memcpy(wavetemp, wavetable[c], MAX_WAVELEN*2);

    for (d = 0; d < MAX_WAVELEN; d++)
    {
      if (wavetemp[d*2] == 0xff)
      {
        break;
      }
    }
    instrwavesize = d*2 + 2;

    if (wavetblsize < instrwavesize)
    {
      memcpy(&wavetblwork[wavetblsize], wavetemp, instrwavesize);
      instrwork[(c-1)*8+7] = wavetblsize/2 + 1;
      wavetblsize += instrwavesize;
    }
    else
    {
      int duplicate = 0;
      /* Search for duplicate from the bytes already in the wavetable */
      for (d = 0; d <= wavetblsize-instrwavesize; d++)
      {
        if (!memcmp(&wavetblwork[d], wavetemp, instrwavesize))
        {
          duplicate = 1;
          instrwork[(c-1)*8+7] = d/2 + 1;
          break;
        }
      }
      if (!duplicate)
      {
        memcpy(&wavetblwork[wavetblsize], wavetemp, instrwavesize);
        instrwork[(c-1)*8+7] = wavetblsize/2 + 1;
        wavetblsize += instrwavesize;
      }
    }
  }
  if (wavetblsize > 512)
  {
    clearscreen();
    printtextc(15, 15, "WAVETABLE EXCEEDS 512 BYTES - SONG WOULDN'T PLAY PROPERLY ON C64");
    fliptoscreen();
    waitkey2();
    goto PRCLEANUP;
  }

  playerhandle = io_open("player.bin");
  if (playerhandle == -1)
  {
    clearscreen();
    printtextc(15, 15, "CANNOT READ PLAYER.BIN FROM DATAFILE");
    fliptoscreen();
    waitkey2();
    goto PRCLEANUP;
  }

  playersize = io_lseek(playerhandle, 0, SEEK_END);
  playerwork = malloc(playersize);
  if (!playerwork)
  {
    io_close(playerhandle);
    clearscreen();
    printtextc(15, 15, "OUT OF MEMORY IN PACKER/RELOCATOR!");
    fliptoscreen();
    waitkey2();
    goto PRCLEANUP;
  }
  io_lseek(playerhandle, 0, SEEK_SET);
  io_read(playerhandle, playerwork, playersize);
  io_close(playerhandle);

  totalsize = playersize+songtblsize+songsize+patttblsize+pattsize+instrsize+wavetblsize;

  printblank(0, 15, 80);

  sprintf(textbuffer, "PACKING RESULTS:");
  printtext(1, 2, 15, textbuffer);
  sprintf(textbuffer, "Playroutine:     %d bytes", playersize);
  printtext(1, 4, 7, textbuffer);
  sprintf(textbuffer, "Songtable:       %d bytes", songtblsize);
  printtext(1, 5, 7, textbuffer);
  sprintf(textbuffer, "Song-orderlists: %d bytes", songsize);
  printtext(1, 6, 7, textbuffer);
  sprintf(textbuffer, "Patterntable:    %d bytes", patttblsize);
  printtext(1, 7, 7, textbuffer);
  sprintf(textbuffer, "Patterns:        %d bytes", pattsize);
  printtext(1, 8, 7, textbuffer);
  sprintf(textbuffer, "Instruments:     %d bytes", instrsize);
  printtext(1, 9, 7, textbuffer);
  sprintf(textbuffer, "Wavetables:      %d bytes", wavetblsize);
  printtext(1, 10, 7, textbuffer);
  sprintf(textbuffer, "Total size:      %d bytes", totalsize);
  printtext(1, 12, 7, textbuffer);
  fliptoscreen();

  sprintf(textbuffer, "SELECT START ADDRESS: (CURSORS=MOVE, ENTER=SELECT, ESC=CANCEL)");
  printtext(1, 14, 15, textbuffer);

  while (!selectdone)
  {
    int badaddr = 0;

    if (testoverlap(playeradr, totalsize, 0x100, 0x100))
    {
      sprintf(textbuffer, "$%04X Overlaps stack!            ", playeradr);
      printtext(1, 15, 14, textbuffer);
      badaddr = 1;
    }
    if (testoverlap(playeradr, totalsize, 0x0, 0x100))
    {
      sprintf(textbuffer, "$%04X Overlaps zeropage!         ", playeradr);
      printtext(1, 15, 14, textbuffer);
      badaddr = 1;
    }

    if (testoverlap(playeradr, totalsize, 0xd000, 0x1000))
    {
      sprintf(textbuffer, "$%04X Overlaps I/O area!         ", playeradr);
      printtext(1, 15, 14, textbuffer);
      badaddr = 1;
    }
    if (testoverlap(playeradr, totalsize, 0xfffa, 0x6))
    {
      sprintf(textbuffer, "$%04X Overlaps interrupt vectors!", playeradr);
      printtext(1, 15, 14, textbuffer);
      badaddr = 1;
    }
    if (playeradr + totalsize > 0x10000)
    {
      sprintf(textbuffer, "$%04X Goes past end of memory!   ", playeradr);
      printtext(1, 15, 14, textbuffer);
      badaddr = 1;
    }
    if (!badaddr)
    {
      sprintf(textbuffer, "$%04X Ok address.                ", playeradr);
      printtext(1, 15, 10, textbuffer);
    }
    fliptoscreen();
    waitkey2();

    switch(virtualkey)
    {
      case VK_LEFT:
      playeradr -= 0x0400;
      playeradr &= 0xff00;
      break;

      case VK_UP:
      playeradr += 0x0100;
      playeradr &= 0xff00;
      break;

      case VK_RIGHT:
      playeradr += 0x0400;
      playeradr &= 0xff00;
      break;

      case VK_DOWN:
      playeradr -= 0x0100;
      playeradr &= 0xff00;
      break;

      case VK_ESCAPE:
      selectdone = -1;
      break;

      case VK_RETURN:
      if (!badaddr) selectdone = 1;
      break;
    }
  }

  if (selectdone == -1) goto PRCLEANUP;

  instradr = playeradr + playersize;
  wavetbladr = instradr + instrsize;
  songtbladr = wavetbladr + wavetblsize;
  songadr = songtbladr + songtblsize;
  patttbladr = songadr + songsize;
  pattadr = patttbladr + patttblsize;

  /* Relocate song & patterntables now that all addresses are known */
  for (c = 0; c < songs; c++)
  {
    int adr;

    adr = songtblwork[c*3] | (songtblwork[(c+songs)*3] << 8);
    adr += songadr;
    songtblwork[c*3] = adr & 0xff;
    songtblwork[(c+songs)*3] = adr >> 8;

    adr = songtblwork[c*3+1] | (songtblwork[(c+songs)*3+1] << 8);
    adr += songadr;
    songtblwork[c*3+1] = adr & 0xff;
    songtblwork[(c+songs)*3+1] = adr >> 8;

    adr = songtblwork[c*3+2] | (songtblwork[(c+songs)*3+2] << 8);
    adr += songadr;
    songtblwork[c*3+2] = adr & 0xff;
    songtblwork[(c+songs)*3+2] = adr >> 8;
  }
  for (c = 0; c < patterns; c++)
  {
    int adr;

    adr = patttblwork[c] | (patttblwork[c+patterns] << 8);
    adr += pattadr;
    patttblwork[c] = adr & 0xff;
    patttblwork[c+patterns] = adr >> 8;
  }

  /* Now relocate player */

  c = 0;
  for (;;)
  {
    INSTRUCTION *opptr;
    unsigned char opcode;

    opcode = playerwork[c];
    if (opcode == 0x01) break; /* Reached data, relocation complete */
    opptr = &asmtable[0];

    for (;;)
    {
      if (opptr->opcode == 0xff)
      {
        clearscreen();
        printblankc(0, 0, 15+16, 80);
        printtext(0, 0, 15+16, "GoatTracker PACKER/RELOCATOR");
        printtextc(15, 15, "ILLEGAL OPCODE FOUND IN PLAYROUTINE!");
        fliptoscreen();
        waitkey2();
        goto PRCLEANUP;
      }
      if (opptr->opcode == opcode) break;
      opptr++;
    }

    /* 3 bytes long instructions need relocation */
    if (opptr->length == 3)
    {
      unsigned short *adr = (unsigned short *)&playerwork[c+1];

      switch(playerwork[c+2])
      {
        /* Player code */
        default:
        *adr -= 0x1000;
        *adr += playeradr;
        break;

        /* Reference to instrument data */
        case 0x40:
        *adr -= 0x4000;
        *adr += instradr - 8; /* First tableindex player uses is 8 */
        break;

        /* Reference to wavetable data */
        case 0x50:
        *adr -= 0x5000;
        *adr += wavetbladr - 1;
        break;

        /* Reference to notetable data */
        case 0x51:
        *adr -= 0x5100;
        *adr += wavetbladr+wavetblsize/2 - 1;
        break;

        /* Reference to songtable, lowbytes */
        case 0x60:
        *adr -= 0x6000;
        *adr += songtbladr - 3;
        break;

        /* Reference to songtable, highbytes */
        case 0x70:
        *adr -= 0x7000;
        *adr += songtbladr - 3 + songs*3;
        break;

        /* Reference to patterntable, lowbytes */
        case 0x80:
        *adr -= 0x8000;
        *adr += patttbladr;
        break;

        /* Reference to patterntable, highbytes */
        case 0x90:
        *adr -= 0x9000;
        *adr += patttbladr + patterns;
        break;

        /* Reference to SID registers */
        case 0xd4:
        break;
      }
    }
    c += opptr->length;
  }

  /* Now ask for fileformat */

  printtext(1, 17, 15, "SELECT FORMAT TO SAVE IN: (CURSORS=MOVE, ENTER=SELECT, ESC=CANCEL)");

  selectdone = 0;
  while (!selectdone)
  {
    switch(fileformat)
    {
      case FORMAT_SID:
      printtext(1, 18, 10, "SID - SIDPlay music file format          ");
      strcpy(packedfilter, "*.sid");
      break;

      case FORMAT_PRG:
      printtext(1, 18, 10, "PRG - C64 native format                  ");
      strcpy(packedfilter, "*.prg");
      break;

      case FORMAT_BIN:
      printtext(1, 18, 10, "BIN - Raw binary format (no startaddress)");
      strcpy(packedfilter, "*.bin");
      break;
    }

    fliptoscreen();
    waitkey2();

    switch(virtualkey)
    {
      case VK_LEFT:
      case VK_DOWN:
      fileformat--;
      if (fileformat < FORMAT_SID) fileformat = FORMAT_BIN;
      break;

      case VK_RIGHT:
      case VK_UP:
      fileformat++;
      if (fileformat > FORMAT_BIN) fileformat = FORMAT_SID;
      break;

      case VK_ESCAPE:
      selectdone = -1;
      break;

      case VK_RETURN:
      selectdone = 1;
      break;
    }
  }
  if (selectdone == -1) goto PRCLEANUP;

  memset(packedsongname, 0, sizeof packedsongname);

  /* Now ask for filename */
  if (!fileselector(packedsongname, packedfilter, "Save Music+Playroutine", 2))
    goto PRCLEANUP;

  songhandle = open(packedsongname, O_RDWR | O_BINARY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
  if (songhandle != -1)
  {
    if (fileformat == FORMAT_PRG)
    {
      write(songhandle, &playeradr, 2);
    }
    if (fileformat == FORMAT_SID)
    {
      unsigned char ident[] = {'P', 'S', 'I', 'D', 0x00, 0x01, 0x00, 0x76};
      unsigned char initrout[] = {0xaa, 0xe8, 0x8e, 0x00, 0x00, 0x60};
      unsigned char byte;

      /* Identification */
      write(songhandle, ident, sizeof ident);

      /* Load address */
      byte = (playeradr-6) >> 8;
      write(songhandle, &byte, sizeof byte);
      byte = (playeradr-6) & 0xff;
      write(songhandle, &byte, sizeof byte);

      /* Init address */
      byte = (playeradr-6) >> 8;
      write(songhandle, &byte, sizeof byte);
      byte = (playeradr-6) & 0xff;
      write(songhandle, &byte, sizeof byte);

      /* Play address */
      byte = (playeradr) >> 8;
      write(songhandle, &byte, sizeof byte);
      byte = (playeradr) & 0xff;
      write(songhandle, &byte, sizeof byte);

      /* Number of subtunes */
      byte = 0x00;
      write(songhandle, &byte, sizeof byte);
      byte = songs;
      write(songhandle, &byte, sizeof byte);

      /* Default subtune */
      byte = 0x00;
      write(songhandle, &byte, sizeof byte);
      byte = 0x01;
      write(songhandle, &byte, sizeof byte);

      /* Song speed bits */
      byte = 0x00;
      write(songhandle, &byte, sizeof byte);
      write(songhandle, &byte, sizeof byte);
      write(songhandle, &byte, sizeof byte);
      write(songhandle, &byte, sizeof byte);

      /* Songname etc. */
      write(songhandle, songname, sizeof songname);
      write(songhandle, authorname, sizeof authorname);
      write(songhandle, copyrightname, sizeof copyrightname);

      /* Init routine */
      initrout[3] = (playeradr+1) & 0xff;
      initrout[4] = (playeradr+1) >> 8;
      write(songhandle, initrout, sizeof initrout);
    }

    write(songhandle, playerwork, playersize);
    write(songhandle, instrwork, instrsize);
    for (c = 0; c < wavetblsize/2; c++)
    {
      write(songhandle, &wavetblwork[c*2], 1);
    }
    for (c = 0; c < wavetblsize/2; c++)
    {
      write(songhandle, &wavetblwork[c*2+1], 1);
    }
    write(songhandle, songtblwork, songtblsize);
    write(songhandle, songwork, songsize);
    write(songhandle, patttblwork, patttblsize);
    write(songhandle, pattwork, pattsize);
    close(songhandle);
  }

  PRCLEANUP:
  if (pattwork) free(pattwork);
  if (patttblwork) free(patttblwork);
  if (songwork) free(songwork);
  if (songtblwork) free(songtblwork);
  if (instrwork) free(instrwork);
  if (wavetblwork) free(wavetblwork);
  if (playerwork) free(playerwork);
  printmainscreen();
}

int packpattern(unsigned char *dest, unsigned char *src, int rows)
{
  unsigned char temp1[256];
  unsigned char temp2[256];
  int count;
  unsigned char instr;
  int command;
  int databyte;
  int destsizeim;
  int destsize;

  /*
   * First write the pattern in such format that zero instruments are
   * eliminated
   */
  for (count = 0; count < rows; count++)
  {
    if (!count)
    {
      temp1[count*3] = src[count*3];
      temp1[count*3+1] = src[count*3+1];
      temp1[count*3+2] = src[count*3+2];
      instr = src[count*3+1] & 0xf8;
      if (instr > highestusedinstr) highestusedinstr = instr;
    }
    else
    {
      if (src[count*3+1] & 0xf8)
      {
        temp1[count*3] = src[count*3];
        temp1[count*3+1] = src[count*3+1];
        temp1[count*3+2] = src[count*3+2];
        instr = src[count*3+1] & 0xf8;
        if (instr > highestusedinstr) highestusedinstr = instr;
      }
      else
      {
        temp1[count*3] = src[count*3];
        temp1[count*3+1] = src[count*3+1] | instr;
        temp1[count*3+2] = src[count*3+2];
      }
    }
    /* Decrease databyte of all tempo commands for the playroutine */
    if ((temp1[count*3+1] & 0x07) == CMD_SETTEMPO)
    {
      temp1[count*3+2]--;
    }
    /* Swap nybbles of vibrato command for the playroutine */
    if ((temp1[count*3+1] & 0x07) == CMD_VIBRATO)
    {
      temp1[count*3+2] = swapnybbles(temp1[count*3+2]);
    }
  }

  /*
   * Then optimize rows where command & databyte don't change
   */
  command = -1;
  databyte = -1;

  destsizeim = 0;
  for (count = 0; count < rows; count++)
  {
    if ((temp1[count*3+1] != command) ||
        (temp1[count*3+2] != databyte))
    {
      command = temp1[count*3+1];
      databyte = temp1[count*3+2];
      temp2[destsizeim++] = temp1[count*3];
      temp2[destsizeim++] = command;
      temp2[destsizeim++] = databyte;
    }
    else
    {
      temp2[destsizeim++] = temp1[count*3] + 0x60;
    }
  }

  /*
   * Then optimize long singlebyte rests with "packed rest" (final phase)
   */
  destsize = 0;
  for (count = 0; count < destsizeim;)
  {
    if (temp2[count] < 0x60)
    {
      dest[destsize++] = temp2[count++];
      dest[destsize++] = temp2[count++];
      dest[destsize++] = temp2[count++];
    }
    else
    {
      /* Never pack the last row with a packed rest */
      if ((temp2[count] != 0x60+REST) || (count >= destsizeim - 1))
      {
        dest[destsize++] = temp2[count++];
      }
      else
      {
        int d;
        for (d = count; d < destsizeim-1; )
        {
          if (temp2[d] == 0x60+REST)
          {
            d++;
            if (d - count == 64) break;
          }
          else break;
        }
        d -= count;

        if (d > 1)
        {
          dest[destsize++] = -d;
          count += d;
        }
        else
        {
          dest[destsize++] = temp2[count++];
        }
      }
    }
  }
  dest[destsize++] = ENDPATT;

  return destsize;
}

int testoverlap(int area1start, int area1size, int area2start, int area2size)
{
  int area1last = area1start+area1size-1;
  int area2last = area2start+area2size-1;

  if (area1start == area2start) return 1;

  if (area1start < area2start)
  {
    if (area1last < area2start) return 0;
    else return 1;
  }
  else
  {
    if (area2last < area1start) return 0;
    else return 1;
  }
}
